from typing import List

from supervisely import Api, Project
from supervisely.app.widgets import Button, Card, Container, Text, TrainValSplits


class TrainValSplitsSelector:
    title = "Train / Val Splits"
    description = "Select train and val splits for training"
    lock_message = "Select input options to unlock"

    def __init__(self, api: Api, project_id: int, app_options: dict = {}):
        # Init widgets
        self.train_val_splits = None
        self.validator_text = None
        self.button = None
        self.container = None
        self.card = None
        # -------------------------------- #

        self.display_widgets = []
        self.app_options = app_options
        self.api = api
        self.project_id = project_id

        # GUI Components
        split_methods = self.app_options.get("train_val_split_methods", [])
        if len(split_methods) == 0:
            split_methods = ["Random", "Based on tags", "Based on datasets", "Based on collections"]
        random_split = "Random" in split_methods
        tag_split = "Based on tags" in split_methods
        ds_split = "Based on datasets" in split_methods
        coll_split = "Based on collections" in split_methods

        self.train_val_splits = TrainValSplits(
            project_id, None, random_split, tag_split, ds_split, collections_splits=coll_split
        )

        # check for collections with "train" and "val" prefixes
        all_collections = self.api.entities_collection.get_list(self.project_id)
        train_collections = []
        val_collections = []
        collections_found = False
        for collection in all_collections:
            if collection.name.lower().startswith("train"):
                train_collections.append(collection.id)
            elif collection.name.lower().startswith("val"):
                val_collections.append(collection.id)
        
        if len(train_collections) > 0 and len(val_collections) > 0:
            self.train_val_splits.set_collections_splits(train_collections, val_collections)
            self.validator_text = Text(
                "Train and val collections are detected", status="info"
            )
            self.validator_text.show()
            collections_found = True
        else:
            self.validator_text = Text("")
            self.validator_text.hide()
        

        def _extend_with_nested(root_ds):
            nested = self.api.dataset.get_nested(self.project_id, root_ds.id)
            nested_ids = [ds.id for ds in nested]
            return [root_ds.id] + nested_ids

        if not collections_found:
            train_val_dataset_ids = {"train": set(), "val": set()}
            for _, dataset in self.api.dataset.tree(self.project_id):
                ds_name = dataset.name.lower()

                if ds_name in {"train", "training"}:
                    for _id in _extend_with_nested(dataset):
                        train_val_dataset_ids["train"].add(_id)

                elif ds_name in {"val", "validation", "test", "testing"}:
                    for _id in _extend_with_nested(dataset):
                        train_val_dataset_ids["val"].add(_id)

            train_val_dataset_ids["train"] = list(train_val_dataset_ids["train"])
            train_val_dataset_ids["val"] = list(train_val_dataset_ids["val"])

            train_count = len(train_val_dataset_ids["train"])
            val_count = len(train_val_dataset_ids["val"])

            if train_count > 0 and val_count > 0:
                self.train_val_splits.set_datasets_splits(
                    train_val_dataset_ids["train"], train_val_dataset_ids["val"]
                )

            if train_count > 0 and val_count > 0:
                if train_count == val_count == 1:
                    self.validator_text = Text("train and val datasets are detected", status="info")
                else:
                    self.validator_text = Text(
                        "Multiple train and val datasets are detected. Check manually if selection is correct",
                        status="info",
                    )
                self.validator_text.show()
            else:
                self.validator_text = Text("")
                self.validator_text.hide()

        self.button = Button("Select")
        self.display_widgets.extend([self.train_val_splits, self.validator_text, self.button])
        # -------------------------------- #

        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
            collapsable=self.app_options.get("collapsable", False),
        )
        self.card.lock()

    @property
    def widgets_to_disable(self) -> list:
        return [self.train_val_splits]

    def validate_step(self) -> bool:
        split_method = self.train_val_splits.get_split_method()
        warning_text = "Using the same data for training and validation leads to overfitting, poor generalization and biased model selection."
        ensure_text = "Ensure this is intentional."
        is_valid = False

        def validate_random_split():
            train_ratio = self.train_val_splits.get_train_split_percent()
            val_ratio = self.train_val_splits.get_val_split_percent()

            # Define common warning text
            ensure_text_random_split = (
                "Consider reallocating to ensure efficient learning and validation."
            )

            # Validate splits
            if train_ratio == val_ratio:
                self.validator_text.set(
                    text="Train and validation splits are equal (50:50). This is inefficient for standard training. "
                    f"{ensure_text}",
                    status="warning",
                )
            elif train_ratio > 90:
                self.validator_text.set(
                    text="Training split exceeds 90%. This may leave insufficient data for validation. Ensure you have enough data for validation.",
                    status="warning",
                )
            elif val_ratio > train_ratio:
                self.validator_text.set(
                    text=f"Validation split is larger than the training split. {ensure_text_random_split}",
                    status="warning",
                )
            elif train_ratio < 70:
                self.validator_text.set(
                    text="Training split is below 70%. This may limit the model's learning capability. "
                    f"{ensure_text_random_split}",
                    status="warning",
                )
            else:
                self.validator_text.set(
                    text="Train and validation splits are selected.",
                    status="success",
                )
            return True

        def validate_based_on_tags():
            train_tag = self.train_val_splits.get_train_tag()
            val_tag = self.train_val_splits.get_val_tag()

            # Check if tags are present on any item in the project
            stats = self.api.project.get_stats(self.project_id)
            tags_count = {}
            for item in stats["imageTags"]["items"]:
                tag_name = item["tagMeta"]["name"]
                tag_total = item["total"]
                tags_count[tag_name] = tag_total

            for object_tags in stats["objectTags"]["items"]:
                tag_name = object_tags["tagMeta"]["name"]
                tag_total = object_tags["total"]
                if tag_name in tags_count:
                    tags_count[tag_name] += tag_total
                else:
                    tags_count[tag_name] = tag_total

            if tags_count[train_tag] == 0:
                self.validator_text.set(
                    text=f"Train tag '{train_tag}' is not present in any images. {ensure_text}",
                    status="error",
                )
                return False
            elif tags_count[val_tag] == 0:
                self.validator_text.set(
                    text=f"Val tag '{val_tag}' is not present in any images. {ensure_text}",
                    status="error",
                )
                return False
            elif train_tag == val_tag:
                self.validator_text.set(
                    text=f"Train and val tags are the same. {ensure_text} {warning_text}",
                    status="warning",
                )
                return True
            else:
                self.validator_text.set("Train and val tags are selected", status="success")
                return True

        def validate_based_on_datasets():
            train_dataset_id = self.get_train_dataset_ids()
            val_dataset_id = self.get_val_dataset_ids()
            if train_dataset_id is None and val_dataset_id is None:
                self.validator_text.set("No datasets are selected", status="error")
                return False

            if train_dataset_id is None:
                self.validator_text.set("No train dataset is selected", status="error")
                return False

            if val_dataset_id is None:
                self.validator_text.set("No val dataset is selected", status="error")
                return False

            # Check if datasets are not empty
            stats = self.api.project.get_stats(self.project_id)
            datasets_count = {}
            for dataset in stats["images"]["datasets"]:
                datasets_count[dataset["id"]] = {
                    "name": dataset["name"],
                    "total": dataset["imagesInDataset"],
                }

            empty_dataset_names = []
            for dataset_id in train_dataset_id + val_dataset_id:
                if datasets_count[dataset_id]["total"] == 0:
                    empty_dataset_names.append(datasets_count[dataset_id]["name"])

            if len(empty_dataset_names) > 0:
                if len(empty_dataset_names) == len(train_dataset_id + val_dataset_id):
                    empty_ds_text = f"All selected datasets are empty. {ensure_text}"
                    self.validator_text.set(
                        text=empty_ds_text,
                        status="error",
                    )
                    return False

                if len(empty_dataset_names) == 1:
                    empty_ds_text = f"Selected dataset: {', '.join(empty_dataset_names)} is empty. {ensure_text}"
                else:
                    empty_ds_text = f"Selected datasets: {', '.join(empty_dataset_names)} are empty. {ensure_text}"

                self.validator_text.set(
                    text=empty_ds_text,
                    status="error",
                )
                return True

            elif train_dataset_id == val_dataset_id:
                self.validator_text.set(
                    text=f"Same datasets are selected for both train and val splits. {ensure_text} {warning_text}",
                    status="warning",
                )
                return True
            else:
                self.validator_text.set("Train and val datasets are selected", status="success")
                return True

        def validate_based_on_collections():
            train_collection_id = self.train_val_splits.get_train_collections_ids()
            val_collection_id = self.train_val_splits.get_val_collections_ids()
            if train_collection_id is None and val_collection_id is None:
                self.validator_text.set("No collections are selected", status="error")
                return False
            if len(train_collection_id) == 0 or len(val_collection_id) == 0:
                self.validator_text.set("Collections are not selected", status="error")
                return False
            if set(train_collection_id) == set(val_collection_id):
                self.validator_text.set(
                    text=f"Same collections are selected for both train and val splits. {ensure_text} {warning_text}",
                    status="warning",
                )
                return True
            from supervisely.api.entities_collection_api import CollectionTypeFilter

            train_items = set()
            empty_train_collections = []
            for collection_id in train_collection_id:
                items = self.api.entities_collection.get_items(
                    collection_id=collection_id,
                    project_id=self.project_id,
                    collection_type=CollectionTypeFilter.DEFAULT,
                )
                train_items.update([item.id for item in items])
                if len(items) == 0:
                    empty_train_collections.append(collection_id)
            val_items = set()
            empty_val_collections = []
            for collection_id in val_collection_id:
                items = self.api.entities_collection.get_items(
                    collection_id=collection_id,
                    project_id=self.project_id,
                    collection_type=CollectionTypeFilter.DEFAULT,
                )
                val_items.update([item.id for item in items])
                if len(items) == 0:
                    empty_val_collections.append(collection_id)
            if len(train_items) == 0 and len(val_items) == 0:
                self.validator_text.set(
                    text="All selected collections are empty. ",
                    status="error",
                )
                return False
            if len(empty_train_collections) > 0 or len(empty_val_collections) > 0:
                empty_collections_text = "Selected collections are empty. "
                if len(empty_train_collections) > 0:
                    empty_collections_text += f"train: {', '.join(empty_train_collections)}. "
                if len(empty_val_collections) > 0:
                    empty_collections_text += f"val: {', '.join(empty_val_collections)}. "
                empty_collections_text += f"{ensure_text}"
                self.validator_text.set(
                    text=empty_collections_text,
                    status="error",
                )
                return True

            else:
                self.validator_text.set("Train and val collections are selected", status="success")
                return True

        if split_method == "Random":
            is_valid = validate_random_split()

        elif split_method == "Based on tags":
            is_valid = validate_based_on_tags()

        elif split_method == "Based on datasets":
            is_valid = validate_based_on_datasets()
        elif split_method == "Based on collections":
            is_valid = validate_based_on_collections()

        # @TODO: handle button correctly if validation fails. Do not unlock next card until validation passes if returned False
        self.validator_text.show()
        return is_valid

    def set_sly_project(self, project: Project) -> None:
        self.train_val_splits._project_fs = project

    def get_split_method(self) -> str:
        return self.train_val_splits.get_split_method()

    def get_train_dataset_ids(self) -> List[int]:
        return self.train_val_splits._train_ds_select.get_selected_ids()

    def set_train_dataset_ids(self, dataset_ids: List[int]) -> None:
        self.train_val_splits._train_ds_select.set_selected_ids(dataset_ids)

    def get_val_dataset_ids(self) -> List[int]:
        return self.train_val_splits._val_ds_select.get_selected_ids()

    def set_val_dataset_ids(self, dataset_ids: List[int]) -> None:
        self.train_val_splits._val_ds_select.set_selected_ids(dataset_ids)

    def get_train_collection_ids(self) -> List[int]:
        return self.train_val_splits._train_collections_select.get_selected_ids()

    def set_train_collection_ids(self, collection_ids: List[int]) -> None:
        self.train_val_splits._train_collections_select.set_selected_ids(collection_ids)

    def get_val_collection_ids(self) -> List[int]:
        return self.train_val_splits._val_collections_select.get_selected_ids()

    def set_val_collection_ids(self, collection_ids: List[int]) -> None:
        self.train_val_splits._val_collections_select.set_selected_ids(collection_ids)



import supervisely.nn.benchmark.instance_segmentation.text_templates as vis_texts
from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.nn.benchmark.object_detection.visualizer import (
    ObjectDetectionVisualizer,
)


class InstanceSegmentationVisualizer(ObjectDetectionVisualizer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.vis_texts = vis_texts
        self._widgets = False
        self.ann_opacity = 0.7

    @property
    def cv_task(self):
        return CVTask.INSTANCE_SEGMENTATION



from typing import Dict, Optional, Union

from supervisely.api.dataset_api import DatasetInfo
from supervisely.api.project_api import ProjectInfo
from supervisely.app import DataJson
from supervisely.app.widgets import Widget
from supervisely.project.project import Dataset, Project


class DatasetThumbnail(Widget):
    """DatasetThumbnail widget in Supervisely is a widget that allows to display a thumbnail image that represents supervisely dataset.

    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/thumbnails/datasetthumbnail>`_
        (including screenshots and examples).

    :param project_info: project info
    :type project_info: Optional[ProjectInfo]
    :param dataset_info: dataset info
    :type dataset_info: Optional[DatasetInfo]
    :param show_project_name: if True, project name will be shown
    :type show_project_name: Optional[bool]
    :param widget_id: An identifier of the widget.
    :type widget_id: str, optional

    :Usage example:
    .. code-block:: python

            from supervisely.app.widgets import DatasetThumbnail

            dataset_thumbnail = DatasetThumbnail(
                project_info=project_info,
                dataset_info=dataset_info,
            )
    """

    def __init__(
        self,
        project_info: Optional[ProjectInfo] = None,
        dataset_info: Optional[DatasetInfo] = None,
        show_project_name: Optional[bool] = True,
        widget_id: Optional[str] = None,
    ):
        self._project_info: ProjectInfo = None
        self._dataset_info: DatasetInfo = None
        self._id: int = None
        self._name: str = None
        self._description: str = None
        self._url: str = None
        self._image_preview_url: str = None
        self._show_project_name: bool = show_project_name
        self._project_name: str = None
        self._project_url: str = None
        self._set_info(project_info, dataset_info, show_project_name)

        super().__init__(widget_id=widget_id, file_path=__file__)

    def get_json_data(self) -> Dict[str, Union[int, str, bool]]:
        """Returns dictionary with widget data, which defines the appearance and behavior of the widget.

        Dictionary contains the following fields:
            - id: dataset id
            - name: dataset name
            - description: dataset description
            - url: dataset url
            - image_preview_url: dataset image preview url
            - show_project_name: if True, project name will be shown
            - project_name: project name
            - project_url: project url

        :return: dictionary with widget data
        :rtype: Dict[str, Union[int, str, bool]]
        """
        return {
            "id": self._id,
            "name": self._name,
            "description": self._description,
            "url": self._url,
            "image_preview_url": self._image_preview_url,
            "show_project_name": self._show_project_name,
            "project_name": self._project_name,
            "project_url": self._project_url,
        }

    def get_json_state(self) -> None:
        """DatasetThumbnail widget doesn't have state and returns None."""
        return None

    def _set_info(
        self, project_info: ProjectInfo, dataset_info: DatasetInfo, show_project_name: bool
    ):
        if project_info is None:
            return
        if dataset_info is None:
            return

        self._project_info = project_info
        self._dataset_info = dataset_info
        self._id = dataset_info.id
        self._name = dataset_info.name
        self._description = f"{self._dataset_info.items_count} {self._project_info.type} in dataset"
        self._url = Dataset.get_url(project_id=project_info.id, dataset_id=dataset_info.id)
        self._image_preview_url = dataset_info.image_preview_url
        self._show_project_name = show_project_name
        self._project_name = project_info.name
        self._project_url = Project.get_url(project_info.id)

    def set(
        self,
        project_info: ProjectInfo,
        dataset_info: DatasetInfo,
        show_project_name: Optional[bool] = True,
    ):
        """Sets the data for the widget.

        :param project_info: project info
        :type project_info: ProjectInfo
        :param dataset_info: dataset info
        :type dataset_info: DatasetInfo
        :param show_project_name: if True, project name will be shown
        :type show_project_name: Optional[bool]
        """
        self._set_info(project_info, dataset_info, show_project_name)
        self.update_data()
        DataJson().send_changes()



from typing import Any, Callable

from supervisely.nn.model.model_api import ModelAPI
from supervisely.nn.tracking.boxmot import apply_boxmot
from supervisely.video_annotation.video_annotation import VideoAnnotation


def _get_apply_fn(tracker: Any) -> Callable:
    if tracker.__class__.__module__.startswith("boxmot"):
        return apply_boxmot
    else:
        raise ValueError(
            f"Tracker {tracker.__class__.__module__} is not supported. Please, use boxmot tracker."
        )


def track(video_id: int, tracker, detector: ModelAPI, **kwargs) -> VideoAnnotation:
    apply_fn = _get_apply_fn(tracker)
    if "classes" in kwargs:
        classes = kwargs["classes"]
    else:
        classes = detector.get_classes()
    predictions = detector.predict_detached(video_id=video_id, **kwargs)
    return apply_fn(tracker, predictions, classes)


import os
from typing import List

import supervisely.convert.pointcloud.las.las_helper as las_helper
from supervisely import PointcloudAnnotation, ProjectMeta, logger
from supervisely.convert.base_converter import AvailablePointcloudConverters
from supervisely.convert.pointcloud.pointcloud_converter import PointcloudConverter
from supervisely.io.fs import JUNK_FILES, get_file_ext


class LasConverter(PointcloudConverter):

    def __str__(self) -> str:
        return AvailablePointcloudConverters.LAS

    def validate_format(self) -> bool:
        las_list = []
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                if file in JUNK_FILES:
                    continue
                elif ext in [".las", ".laz"]:
                    las_list.append(full_path)

        # create Items
        self._items = []
        for las_path in las_list:
            ext = get_file_ext(las_path)
            pcd_path = las_path.replace(ext, ".pcd")
            las_helper.las2pcd(las_path, pcd_path)
            if not os.path.exists(pcd_path):
                logger.warn(f"Failed to convert LAS/LAZ to PCD. Skipping: {las_path}")
                continue
            item = self.Item(pcd_path)
            self._items.append(item)
        return self.items_count > 0

    def to_supervisely(
        self,
        item: PointcloudConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> PointcloudAnnotation:
        """Convert to Supervisely format."""
        return item.create_empty_annotation()



import numpy as np

from supervisely.imaging.color import hex2rgb
from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    CollapseWidget,
    MarkdownWidget,
    NotificationWidget,
    TableWidget,
)


class PrCurve(BaseVisMetrics):
    MARKDOWN_PR_CURVE = "markdown_pr_curve"
    MARKDOWN_PR_TRADE_OFFS = "markdown_trade_offs"
    MARKDOWN_WHAT_IS_PR_CURVE = "markdown_what_is_pr_curve"

    @property
    def markdown_widget(self) -> MarkdownWidget:
        text: str = getattr(self.vis_texts, self.MARKDOWN_PR_CURVE).format(
            self.vis_texts.definitions.about_pr_tradeoffs
        )
        return MarkdownWidget(
            name=self.MARKDOWN_PR_CURVE, title="mAP & Precision-Recall Curve", text=text
        )

    @property
    def chart_widget(self) -> ChartWidget:
        return ChartWidget(name="chart_pr_curve", figure=self.get_figure())

    @property
    def collapsed_widget(self) -> CollapseWidget:
        text_pr_trade_offs = getattr(self.vis_texts, self.MARKDOWN_PR_TRADE_OFFS)
        text_pr_curve = getattr(self.vis_texts, self.MARKDOWN_WHAT_IS_PR_CURVE).format(
            self.vis_texts.definitions.confidence_score,
            self.vis_texts.definitions.true_positives,
            self.vis_texts.definitions.false_positives,
        )
        markdown_pr_trade_offs = MarkdownWidget(
            name=self.MARKDOWN_PR_TRADE_OFFS,
            title="About Trade-offs between precision and recall",
            text=text_pr_trade_offs,
        )
        markdown_whatis_pr_curve = MarkdownWidget(
            name=self.MARKDOWN_WHAT_IS_PR_CURVE,
            title="How the PR curve is built?",
            text=text_pr_curve,
        )
        return CollapseWidget(widgets=[markdown_pr_trade_offs, markdown_whatis_pr_curve])

    @property
    def table_widget(self) -> TableWidget:
        res = {}

        columns = [" ", "mAP (0.5:0.95)", "mAP (0.75)"]
        res["content"] = []
        for i, eval_result in enumerate(self.eval_results, 1):
            value_range = round(eval_result.mp.json_metrics()["mAP"], 2)
            value_75 = eval_result.mp.json_metrics()["AP75"] or "-"
            value_75 = round(value_75, 2) if isinstance(value_75, float) else value_75
            model_name = f"[{i}] {eval_result.name}"
            row = [model_name, value_range, value_75]
            dct = {
                "row": row,
                "id": model_name,
                "items": row,
            }
            res["content"].append(dct)

        columns_options = [
            {"disableSort": True},
            {"disableSort": True},
        ]

        res["columns"] = columns
        res["columnsOptions"] = columns_options

        return TableWidget(
            name="table_pr_curve",
            data=res,
            show_header_controls=False,
            # main_column=columns[0],
            fix_columns=1,
        )

    def get_figure(self):  # -> Optional[go.Figure]:
        import plotly.express as px  # pylint: disable=import-error
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        rec_thr = self.eval_results[0].mp.recThrs
        for i, eval_result in enumerate(self.eval_results, 1):
            pr_curve = eval_result.mp.pr_curve().copy()
            pr_curve[pr_curve == -1] = np.nan
            pr_curve = np.nanmean(pr_curve, axis=-1)

            name = f"[{i}] {eval_result.name}"
            color = ",".join(map(str, hex2rgb(eval_result.color))) + ",0.1"
            line = go.Scatter(
                x=eval_result.mp.recThrs,
                y=pr_curve,
                mode="lines",
                name=name,
                fill="tozeroy",
                fillcolor=f"rgba({color})",
                line=dict(color=eval_result.color),
                hovertemplate=name + "<br>Recall: %{x:.2f}<br>Precision: %{y:.2f}<extra></extra>",
                showlegend=True,
            )
            fig.add_trace(line)

        fig.add_trace(
            go.Scatter(
                x=rec_thr,
                y=[1] * len(rec_thr),
                name="Perfect",
                line=dict(color="orange", dash="dash"),
                showlegend=True,
            )
        )
        fig.update_layout(
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
            xaxis_title="Recall",
            yaxis_title="Precision",
        )
        return fig



import inspect
import json
import os
import signal
import sys
from contextlib import suppress
from functools import wraps
from pathlib import Path
from threading import Event as ThreadingEvent
from threading import Thread
from time import sleep
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional

import arel
import jinja2
import numpy as np
import psutil
from async_asgi_testclient import TestClient
from fastapi import (
    Depends,
    FastAPI,
    HTTPException,
    Request,
    Response,
    WebSocket,
    WebSocketDisconnect,
)
from fastapi.exception_handlers import http_exception_handler
from fastapi.responses import JSONResponse
from fastapi.routing import APIRouter
from fastapi.staticfiles import StaticFiles

import supervisely.io.env as sly_env
from supervisely._utils import (
    is_debug_with_sly_net,
    is_development,
    is_docker,
    is_production,
)
from supervisely.api.api import API_TOKEN, SERVER_ADDRESS, TASK_ID, Api
from supervisely.api.module_api import ApiField
from supervisely.app.exceptions import DialogWindowBase
from supervisely.app.fastapi.custom_static_files import CustomStaticFiles
from supervisely.app.fastapi.templating import Jinja2Templates
from supervisely.app.fastapi.utils import run_sync
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.app.singleton import Singleton
from supervisely.app.widgets_context import JinjaWidgets
from supervisely.geometry.bitmap import Bitmap
from supervisely.io.fs import dir_exists, mkdir
from supervisely.sly_logger import create_formatter, logger

# from supervisely.app.fastapi.request import Request

if TYPE_CHECKING:
    from supervisely.app.widgets import Widget

import logging

SUPERVISELY_SERVER_PATH_PREFIX = sly_env.supervisely_server_path_prefix()
if SUPERVISELY_SERVER_PATH_PREFIX and not SUPERVISELY_SERVER_PATH_PREFIX.startswith("/"):
    SUPERVISELY_SERVER_PATH_PREFIX = f"/{SUPERVISELY_SERVER_PATH_PREFIX}"


class ReadyzFilter(logging.Filter):
    def filter(self, record):
        if "/readyz" in record.getMessage() or "/livez" in record.getMessage():
            record.levelno = logging.DEBUG  # Change log level to DEBUG
            record.levelname = "DEBUG"
        return True


def _init_uvicorn_logger():
    uvicorn_logger = logging.getLogger("uvicorn.access")
    for handler in uvicorn_logger.handlers:
        handler.setFormatter(create_formatter())
    uvicorn_logger.addFilter(ReadyzFilter())


_init_uvicorn_logger()


class PrefixRouter(APIRouter):
    def add_api_route(self, path, *args, **kwargs):
        allowed_paths = ["/livez", "/is_alive", "/is_running", "/readyz", "/is_ready"]
        if path in allowed_paths:
            super().add_api_route(path, *args, **kwargs)
        if SUPERVISELY_SERVER_PATH_PREFIX:
            path = SUPERVISELY_SERVER_PATH_PREFIX + path
        super().add_api_route(path, *args, **kwargs)


class Event:
    class Brush:
        class DrawLeftMouseReleased:
            endpoint = "/tools_bitmap_brush_figure_changed"

            def __init__(
                self,
                team_id: int,
                workspace_id: int,
                project_id: int,
                dataset_id: int,
                image_id: int,
                label_id: int,
                object_class_id: int,
                object_class_title: str,
                tool_class_id: int,
                session_id: int,
                tool: str,
                user_id: int,
                job_id: int,
                is_fill: bool,
                is_erase: bool,
                geometry_type: str,
                mask: np.ndarray,
            ):
                self.dataset_id = dataset_id
                self.team_id = team_id
                self.workspace_id = workspace_id
                self.project_id = project_id
                self.image_id = image_id
                self.label_id = label_id
                self.object_class_id = object_class_id
                self.object_class_title = object_class_title
                self.tool_class_id = tool_class_id
                self.session_id = session_id
                self.tool = tool
                self.user_id = user_id
                self.job_id = job_id
                self.is_fill = is_fill
                self.is_erase = is_erase
                self.geometry_type = geometry_type
                self.mask = mask

            @classmethod
            def from_json(cls, data: Dict[str, Any]):
                tool_state = data.get(ApiField.TOOL_STATE)
                if tool_state is not None:
                    tool_option = tool_state.get(ApiField.OPTION)
                    if tool_option == "fill":
                        is_fill = True
                        is_erase = False
                    elif tool_option == "erase":
                        is_fill = False
                        is_erase = True
                    else:
                        raise ValueError(f"Unknown tool option: {tool_option}")
                else:
                    is_fill = False
                    is_erase = False
                geometry_type = None
                mask = None
                figure_state = data.get(ApiField.FIGURE_STATE)
                try:
                    geometry_type = figure_state.get(ApiField.GEOMETRY_TYPE)
                    geometry = figure_state.get(ApiField.GEOMETRY).get(ApiField.BITMAP)
                    mask = Bitmap.base64_2_data(geometry.get(ApiField.DATA))
                except AttributeError:
                    pass
                return cls(
                    team_id=data.get(ApiField.TEAM_ID),
                    workspace_id=data.get(ApiField.WORKSPACE_ID),
                    project_id=data.get(ApiField.PROJECT_ID),
                    dataset_id=data.get(ApiField.DATASET_ID),
                    image_id=data.get(ApiField.IMAGE_ID),
                    label_id=data.get(ApiField.FIGURE_ID),
                    object_class_id=data.get(ApiField.FIGURE_CLASS_ID),
                    object_class_title=data.get(ApiField.FIGURE_CLASS_TITLE),
                    tool_class_id=data.get(ApiField.TOOL_CLASS_ID),
                    session_id=data.get(ApiField.SESSION_ID),
                    tool=data.get(ApiField.LABELING_TOOL),
                    user_id=data.get(ApiField.USER_ID),
                    job_id=data.get(ApiField.JOB_ID),
                    is_fill=is_fill,
                    is_erase=is_erase,
                    geometry_type=geometry_type,
                    mask=mask,
                )

    class ManualSelected:
        class VideoChanged:
            endpoint = "/manual_selected_entity_changed"

            def __init__(
                self,
                dataset_id: int,
                team_id: int,
                workspace_id: int,
                project_id: int,
                figure_id: int,
                video_id: int,
                frame: int,
                tool_class_id: int,
                session_id: str,
                tool: str,
                user_id: int,
                job_id: int,
            ):
                self.dataset_id = dataset_id
                self.team_id = team_id
                self.workspace_id = workspace_id
                self.project_id = project_id
                self.figure_id = figure_id
                self.video_id = video_id
                self.frame = frame
                self.tool_class_id = tool_class_id
                self.session_id = session_id
                self.tool = tool
                self.user_id = user_id
                self.job_id = job_id

            @classmethod
            def from_json(cls, data: Dict[str, Any]):
                return cls(
                    dataset_id=data.get(ApiField.DATASET_ID),
                    team_id=data.get(ApiField.TEAM_ID),
                    workspace_id=data.get(ApiField.WORKSPACE_ID),
                    project_id=data.get(ApiField.PROJECT_ID),
                    figure_id=data.get(ApiField.FIGURE_ID),
                    video_id=data.get(ApiField.ENTITY_ID),
                    frame=data.get(ApiField.FRAME),
                    tool_class_id=data.get(ApiField.TOOL_CLASS_ID),
                    session_id=data.get(ApiField.SESSION_ID),
                    tool=data.get(ApiField.LABELING_TOOL),
                    user_id=data.get(ApiField.USER_ID),
                    job_id=data.get(ApiField.JOB_ID),
                )

        class FigureChanged:
            endpoint = "/manual_selected_figure_changed"

            def __init__(
                self,
                dataset_id: int,
                team_id: int,
                workspace_id: int,
                project_id: int,
                figure_id: int,
                figure_class_id: int,
                figure_class_title: str,
                image_id: int,
                video_id: int,
                frame: int,
                object_id: int,
                tool_class_id: int,
                session_id: str,
                tool: str,
                user_id: int,
                job_id: int,
                previous_figure: dict = None,
            ):
                self.dataset_id = dataset_id
                self.team_id = team_id
                self.workspace_id = workspace_id
                self.project_id = project_id
                self.figure_id = figure_id
                self.image_id = image_id
                self.figure_class_id = figure_class_id
                self.figure_class_title = figure_class_title
                self.video_id = video_id
                self.frame = frame
                self.object_id = object_id
                self.tool_class_id = tool_class_id
                self.session_id = session_id
                self.tool = tool
                self.user_id = user_id
                self.job_id = job_id
                self.previous_figure = previous_figure

            @classmethod
            def from_json(cls, data: dict):
                return cls(
                    dataset_id=data.get(ApiField.DATASET_ID),
                    team_id=data.get(ApiField.TEAM_ID),
                    workspace_id=data.get(ApiField.WORKSPACE_ID),
                    project_id=data.get(ApiField.PROJECT_ID),
                    figure_id=data.get(ApiField.FIGURE_ID),
                    figure_class_id=data.get(ApiField.FIGURE_CLASS_ID),
                    figure_class_title=data.get(ApiField.FIGURE_CLASS_TITLE),
                    image_id=data.get(ApiField.ENTITY_ID),
                    video_id=data.get(ApiField.ENTITY_ID),
                    frame=data.get(ApiField.FRAME),
                    object_id=data.get(ApiField.ANNOTATION_OBJECT_ID),
                    tool_class_id=data.get(ApiField.TOOL_CLASS_ID),
                    session_id=data.get(ApiField.SESSION_ID),
                    tool=data.get(ApiField.LABELING_TOOL),
                    user_id=data.get(ApiField.USER_ID),
                    job_id=data.get(ApiField.JOB_ID),
                    previous_figure=data.get(
                        "previousFigure", None
                    ),  # there is no such field in ApiField
                )

        class ImageChanged:
            """This event is triggered when the user changes the image in the Image Labeling Tooolbox."""

            endpoint = "/manual_selected_image_changed"

            def __init__(
                self,
                dataset_id: int,
                team_id: int,
                workspace_id: int,
                project_id: int,
                image_id: int,
                figure_id: int,
                figure_class_id: int,
                figure_class_title: str,
                tool_class_id: int,
                session_id: str,
                tool: str,
                user_id: int,
                job_id: int,
            ):
                self.dataset_id = dataset_id
                self.team_id = team_id
                self.workspace_id = workspace_id
                self.project_id = project_id
                self.image_id = image_id
                self.figure_id = figure_id
                self.figure_class_id = figure_class_id
                self.figure_class_title = figure_class_title
                self.tool_class_id = tool_class_id
                self.session_id = session_id
                self.tool = tool
                self.user_id = user_id
                self.job_id = job_id

            @classmethod
            def from_json(cls, data: dict):
                return cls(
                    dataset_id=data.get(ApiField.DATASET_ID),
                    team_id=data.get(ApiField.TEAM_ID),
                    workspace_id=data.get(ApiField.WORKSPACE_ID),
                    project_id=data.get(ApiField.PROJECT_ID),
                    image_id=data.get(ApiField.IMAGE_ID),
                    figure_id=data.get(ApiField.FIGURE_ID),
                    figure_class_id=data.get(ApiField.FIGURE_CLASS_ID),
                    figure_class_title=data.get(ApiField.FIGURE_CLASS_TITLE),
                    tool_class_id=data.get(ApiField.TOOL_CLASS_ID),
                    session_id=data.get(ApiField.SESSION_ID),
                    tool=data.get(ApiField.LABELING_TOOL),
                    user_id=data.get(ApiField.USER_ID),
                    job_id=data.get(ApiField.JOB_ID),
                )

    class FigureCreated:
        endpoint = "/figure_created"

        def __init__(
            self,
            dataset_id: int,
            team_id: int,
            workspace_id: int,
            project_id: int,
            figure_id: int,
            figure_class_id: int,
            figure_class_title: str,
            image_id: int,
            video_id: int,
            frame: int,
            object_id: int,
            tool_class_id: int,
            session_id: str,
            tool: str,
            user_id: int,
            job_id: int,
            tool_state: dict,
            figure_state: dict,
        ):
            self.dataset_id = dataset_id
            self.team_id = team_id
            self.workspace_id = workspace_id
            self.project_id = project_id
            self.figure_id = figure_id
            self.figure_class_id = figure_class_id
            self.figure_class_title = figure_class_title
            self.image_id = image_id
            self.video_id = video_id
            self.frame = frame
            self.object_id = object_id
            self.tool_class_id = tool_class_id
            self.session_id = session_id
            self.tool = tool
            self.user_id = user_id
            self.job_id = job_id
            self.tool_state = tool_state
            self.figure_state = figure_state

        @classmethod
        def from_json(cls, data: dict):
            return cls(
                dataset_id=data.get(ApiField.DATASET_ID),
                team_id=data.get(ApiField.TEAM_ID),
                workspace_id=data.get(ApiField.WORKSPACE_ID),
                project_id=data.get(ApiField.PROJECT_ID),
                figure_id=data.get(ApiField.FIGURE_ID),
                figure_class_id=data.get(ApiField.FIGURE_CLASS_ID),
                figure_class_title=data.get(ApiField.FIGURE_CLASS_TITLE),
                image_id=data.get(ApiField.IMAGE_ID),
                video_id=data.get(ApiField.ENTITY_ID),
                frame=data.get(ApiField.FRAME),
                object_id=data.get(ApiField.ANNOTATION_OBJECT_ID),
                tool_class_id=data.get(ApiField.TOOL_CLASS_ID),
                session_id=data.get(ApiField.SESSION_ID),
                tool=data.get(ApiField.LABELING_TOOL),
                user_id=data.get(ApiField.USER_ID),
                job_id=data.get(ApiField.JOB_ID),
                tool_state=data.get(ApiField.JOB_ID),
                figure_state=data.get(ApiField.FIGURE_STATE),
            )

    class Tools:
        class Rectangle:
            class FigureChanged:
                endpoint = "/tools_rectangle_figure_changed"

                def __init__(
                    self,
                    dataset_id: int,
                    team_id: int,
                    workspace_id: int,
                    project_id: int,
                    figure_id: int,
                    figure_class_id: int,
                    figure_class_title: str,
                    image_id: int,
                    tool_class_id: int,
                    session_id: str,
                    tool: str,
                    user_id: int,
                    job_id: int,
                    tool_state: dict,
                    figure_state: dict,
                ):
                    self.dataset_id = dataset_id
                    self.team_id = team_id
                    self.workspace_id = workspace_id
                    self.project_id = project_id
                    self.figure_id = figure_id
                    self.figure_class_id = figure_class_id
                    self.figure_class_title = figure_class_title
                    self.image_id = image_id
                    self.tool_class_id = tool_class_id
                    self.session_id = session_id
                    self.tool = tool
                    self.user_id = user_id
                    self.job_id = job_id
                    self.tool_state = tool_state
                    self.figure_state = figure_state

                @classmethod
                def from_json(cls, data: dict):
                    return cls(
                        dataset_id=data.get(ApiField.DATASET_ID),
                        team_id=data.get(ApiField.TEAM_ID),
                        workspace_id=data.get(ApiField.WORKSPACE_ID),
                        project_id=data.get(ApiField.PROJECT_ID),
                        figure_id=data.get(ApiField.FIGURE_ID),
                        figure_class_id=data.get(ApiField.FIGURE_CLASS_ID),
                        figure_class_title=data.get(ApiField.FIGURE_CLASS_TITLE),
                        image_id=data.get(ApiField.IMAGE_ID),
                        tool_class_id=data.get(ApiField.TOOL_CLASS_ID),
                        session_id=data.get(ApiField.SESSION_ID),
                        tool=data.get(ApiField.LABELING_TOOL),
                        user_id=data.get(ApiField.USER_ID),
                        job_id=data.get(ApiField.JOB_ID),
                        tool_state=data.get(ApiField.JOB_ID),
                        figure_state=data.get(ApiField.FIGURE_STATE),
                    )

    class Entity:
        class FrameChanged:
            endpoint = "/entity_frame_changed"

            def __init__(
                self,
                dataset_id: int,
                team_id: int,
                workspace_id: int,
                project_id: int,
                figure_id: int,
                figure_class_id: int,
                figure_class_title: str,
                video_id: int,
                frame: int,
                object_id: int,
                tool_class_id: int,
                session_id: str,
                tool: str,
                user_id: int,
                job_id: int,
            ):
                self.dataset_id = dataset_id
                self.team_id = team_id
                self.workspace_id = workspace_id
                self.project_id = project_id
                self.figure_id = figure_id
                self.figure_class_id = figure_class_id
                self.figure_class_title = figure_class_title
                self.video_id = video_id
                self.frame = frame
                self.object_id = object_id
                self.tool_class_id = tool_class_id
                self.session_id = session_id
                self.tool = tool
                self.user_id = user_id
                self.job_id = job_id

            @classmethod
            def from_json(cls, data: dict):
                return cls(
                    dataset_id=data.get(ApiField.DATASET_ID),
                    team_id=data.get(ApiField.TEAM_ID),
                    workspace_id=data.get(ApiField.WORKSPACE_ID),
                    project_id=data.get(ApiField.PROJECT_ID),
                    figure_id=data.get(ApiField.FIGURE_ID),
                    figure_class_id=data.get(ApiField.FIGURE_CLASS_ID),
                    figure_class_title=data.get(ApiField.FIGURE_CLASS_TITLE),
                    video_id=data.get(ApiField.ENTITY_ID),
                    frame=data.get(ApiField.FRAME),
                    object_id=data.get(ApiField.ANNOTATION_OBJECT_ID),
                    tool_class_id=data.get(ApiField.TOOL_CLASS_ID),
                    session_id=data.get(ApiField.SESSION_ID),
                    tool=data.get(ApiField.LABELING_TOOL),
                    user_id=data.get(ApiField.USER_ID),
                    job_id=data.get(ApiField.JOB_ID),
                )

    class JobEntity:
        class StatusChanged:
            endpoint = "/job_entity_status_changed"

            def __init__(
                self,
                dataset_id: int,
                team_id: int,
                workspace_id: int,
                project_id: int,
                figure_id: int,
                figure_class_id: int,
                figure_class_title: str,
                image_id: int,
                entity_id: int,
                tool_class_id: int,
                session_id: str,
                tool: str,
                user_id: int,
                job_id: int,
                job_entity_status: str,
            ):
                self.dataset_id = dataset_id
                self.team_id = team_id
                self.workspace_id = workspace_id
                self.project_id = project_id
                self.figure_id = figure_id
                self.figure_class_id = figure_class_id
                self.figure_class_title = figure_class_title
                self.image_id = image_id
                self.entity_id = entity_id
                self.tool_class_id = tool_class_id
                self.session_id = session_id
                self.tool = tool
                self.user_id = user_id
                self.job_id = job_id
                self.job_entity_status = job_entity_status

            @classmethod
            def from_json(cls, data: dict):
                return cls(
                    dataset_id=data.get(ApiField.DATASET_ID),
                    team_id=data.get(ApiField.TEAM_ID),
                    workspace_id=data.get(ApiField.WORKSPACE_ID),
                    project_id=data.get(ApiField.PROJECT_ID),
                    figure_id=data.get(ApiField.FIGURE_ID),
                    figure_class_id=data.get(ApiField.FIGURE_CLASS_ID),
                    figure_class_title=data.get(ApiField.FIGURE_CLASS_TITLE),
                    image_id=data.get(ApiField.IMAGE_ID),
                    entity_id=data.get(ApiField.ENTITY_ID),
                    tool_class_id=data.get(ApiField.TOOL_CLASS_ID),
                    session_id=data.get(ApiField.SESSION_ID),
                    tool=data.get(ApiField.LABELING_TOOL),
                    user_id=data.get(ApiField.USER_ID),
                    job_id=data.get(ApiField.JOB_ID),
                    job_entity_status=data.get(ApiField.JOB_ENTITY_STATUS),
                )


def create(
    process_id=None,
    headless=False,
    auto_widget_id=False,
    before_shutdown_callbacks=None,
) -> FastAPI:
    from supervisely.app import DataJson, StateJson

    JinjaWidgets().auto_widget_id = auto_widget_id
    logger.info(f"JinjaWidgets().auto_widget_id is set to {auto_widget_id}.")

    app = FastAPI()
    WebsocketManager().set_app(app)

    @app.post("/shutdown")
    async def shutdown_endpoint(request: Request):
        shutdown(process_id, before_shutdown_callbacks)

    if headless is False:

        @app.post("/data")
        async def send_data(request: Request):
            data = DataJson()
            response = JSONResponse(content=dict(data))
            return response

        @app.post("/state")
        async def send_state(request: Request):
            state = StateJson()

            response = JSONResponse(content=dict(state))
            gettrace = getattr(sys, "gettrace", None)
            if (gettrace is not None and gettrace()) or is_development():
                response.headers["x-debug-mode"] = "1"
            return response

        @app.post("/session-info")
        async def send_session_info(request: Request):
            # TODO: handle case development inside docker
            production_at_instance = is_production() and is_docker()
            advanced_debug = is_debug_with_sly_net()
            development = is_development() or (is_production() and not is_docker())

            if advanced_debug or development:
                server_address = sly_env.server_address(raise_not_found=False)
                if server_address is not None:
                    server_address = Api.normalize_server_address(server_address)
            elif production_at_instance:
                server_address = "/"
            else:
                raise ValueError(
                    "'Unrecognized running mode, should be one of ['advanced_debug', 'development', 'production']."
                )

            response = JSONResponse(
                content={
                    TASK_ID: os.environ.get(TASK_ID),
                    SERVER_ADDRESS: server_address,
                    API_TOKEN: os.environ.get(API_TOKEN),
                }
            )
            return response

        @app.websocket("/ws")
        async def websocket_endpoint(websocket: WebSocket):
            await WebsocketManager().connect(websocket)
            try:
                while True:
                    data = await websocket.receive_text()
            except WebSocketDisconnect:
                WebsocketManager().disconnect(websocket)

        import supervisely

        app.mount("/css", StaticFiles(directory=supervisely.__path__[0]), name="sly_static")

    return app


def shutdown(
    process_id=None,
    before_shutdown_callbacks: Optional[List[Callable[[], None]]] = None,
):
    logger.info(f"Shutting down [pid argument = {process_id}]...")

    if before_shutdown_callbacks is not None:
        logger.info("Found tasks to run before shutdown.")
        for func in before_shutdown_callbacks:
            logger.debug(f"Call {func.__name__}")
            func()
    else:
        logger.debug("No tasks to call before shutdown")

    try:
        if process_id is None:
            # process_id = psutil.Process(os.getpid()).ppid()
            process_id = os.getpid()
        current_process = psutil.Process(process_id)
        if os.name == "nt":
            # for windows
            current_process.send_signal(signal.CTRL_C_EVENT)  # emit ctrl + c
        else:
            current_process.send_signal(signal.SIGINT)  # emit ctrl + c
    except KeyboardInterrupt:
        logger.info("Application has been shut down successfully")


def enable_hot_reload_on_debug(app: FastAPI):
    templates = Jinja2Templates()
    gettrace = getattr(sys, "gettrace", None)
    if gettrace is None:
        print("Can not detect debug mode, no sys.gettrace")
    elif gettrace():
        # List of directories to exclude from the hot reload.
        exclude = [".venv", ".git", "tmp"]

        hot_reload = arel.HotReload(
            paths=[arel.Path(path) for path in os.listdir() if path not in exclude]
        )

        app.add_websocket_route("/hot-reload", route=hot_reload, name="hot-reload")
        app.add_event_handler("startup", hot_reload.startup)
        app.add_event_handler("shutdown", hot_reload.shutdown)
        templates.env.globals["HOTRELOAD"] = "1"
        templates.env.globals["hot_reload"] = hot_reload
        logger.debug("Debugger (gettrace) detected, UI hot-reload is enabled")
    else:
        logger.debug("In runtime mode ...")


async def process_server_error(request, exc: Exception, need_to_handle_error=True):
    from supervisely.io.exception_handlers import handle_exception

    handled_exception = handle_exception(exc) if need_to_handle_error else None

    if handled_exception is not None:
        details = {"title": handled_exception.title, "message": handled_exception.message}
        log_message = handled_exception.get_message_for_modal_window()
    else:
        details = {"title": "Oops! Something went wrong", "message": repr(exc)}
        log_message = repr(exc)
    if isinstance(exc, DialogWindowBase):
        details["title"] = exc.title
        details["message"] = exc.description
        details["status"] = exc.status

    if is_production():
        logger.error(
            log_message,
            exc_info=True,
            extra={"main_name": "main", "exc_str": details["message"]},
        )
    else:
        raise exc

    return await http_exception_handler(
        request,
        HTTPException(status_code=500, detail=details),
    )


def handle_server_errors(app: FastAPI):
    @app.exception_handler(500)
    async def server_exception_handler(request, exc):
        return await process_server_error(request, exc)


def _init(
    app: FastAPI = None,
    templates_dir: str = "templates",
    headless=False,
    process_id=None,
    static_dir=None,
    hot_reload=False,
    before_shutdown_callbacks=None,
) -> FastAPI:
    from supervisely.app.content import DataJson, StateJson
    from supervisely.app.fastapi import available_after_shutdown

    if app is None:
        app = _MainServer().get_server()

    handle_server_errors(app)

    if headless is False:
        if "app_body_padding" not in StateJson():
            StateJson()["app_body_padding"] = "20px"
        Jinja2Templates(directory=[str(Path(__file__).parent.absolute()), templates_dir])
        if hot_reload:
            enable_hot_reload_on_debug(app)

    StateJson()["slyAppShowDialog"] = False
    DataJson()["slyAppDialogTitle"] = ""
    DataJson()["slyAppDialogMessage"] = ""

    app.mount(
        "/sly",
        create(
            process_id,
            headless,
            auto_widget_id=True,
            before_shutdown_callbacks=before_shutdown_callbacks,
        ),
    )

    @app.middleware("http")
    async def get_state_from_request(request: Request, call_next):
        if headless is False:
            await StateJson.from_request(request)

        if not ("application/json" not in request.headers.get("Content-Type", "")):
            # {'command': 'inference_batch_ids', 'context': {}, 'state': {'dataset_id': 49711, 'batch_ids': [3120204], 'settings': None}, 'user_api_key': 'XXX', 'api_token': 'XXX', 'instance_type': None, 'server_address': 'https://app.supervisely.com'}
            content = await request.json()

            request.state.context = content.get("context")
            request.state.state = content.get("state")
            request.state.api_token = content.get(
                "api_token",
                (
                    request.state.context.get("apiToken")
                    if request.state.context is not None
                    else None
                ),
            )
            # logger.debug(f"middleware request api_token {request.state.api_token}")
            request.state.server_address = content.get(
                "server_address", sly_env.server_address(raise_not_found=False)
            )
            # request.state.server_address = sly_env.server_address(raise_not_found=False)
            # logger.debug(f"middleware request server_address {request.state.server_address}")
            # logger.debug(f"middleware request context {request.state.context}")
            # logger.debug(f"middleware request state {request.state.state}")
            if request.state.server_address is not None and request.state.api_token is not None:
                request.state.api = Api(request.state.server_address, request.state.api_token)
            else:
                request.state.api = None

        try:
            response = await call_next(request)
        except Exception as exc:
            need_to_handle_error = is_production()
            response = await process_server_error(request, exc, need_to_handle_error)
        return response

    def verify_localhost(request: Request):
        client_host = request.client.host
        if client_host not in ["127.0.0.1", "::1"]:
            raise HTTPException(status_code=404, detail="Not Found")

    @app.post("/debug", dependencies=[Depends(verify_localhost)])
    def start_debug():
        import debugpy

        debug_host = os.getenv("DEBUG_HOST", "127.0.0.1")
        debug_port = int(os.getenv("DEBUG_PORT", "5678"))
        debugpy.listen((debug_host, debug_port))
        return {
            "status": "success",
            "message": f"Debug server is listening on {debug_host}:{debug_port}",
            "host": debug_host,
            "port": debug_port,
        }

    if headless is False:
        app.cached_template = None

        @app.get("/")
        @available_after_shutdown(app)
        def read_index(request: Request):
            if request.query_params.get("ping", False) in ("true", "True", True, 1, "1"):
                return JSONResponse(content={"message": "App is running"}, status_code=200)
            if app.cached_template is None:
                app.cached_template = Jinja2Templates().TemplateResponse(
                    "index.html", {"request": request}
                )
            return app.cached_template

        @app.on_event("shutdown")
        def shutdown():
            from supervisely.app.content import ContentOrigin

            ContentOrigin().stop()
            client = TestClient(app)
            resp = run_sync(client.get("/"))
            assert resp.status_code == 200
            logger.info("Application has been shut down successfully")

        if static_dir is not None:
            app.mount("/static", CustomStaticFiles(directory=static_dir), name="static_files")

    return app


class _MainServer(metaclass=Singleton):
    def __init__(self):
        self._server = FastAPI()
        self._server.router = PrefixRouter()

    def get_server(self) -> FastAPI:
        return self._server


class Application(metaclass=Singleton):
    class StopException(Exception):
        """Raise to stop the function from running in app.handle_stop"""

    def __init__(
        self,
        layout: "Widget" = None,
        templates_dir: Optional[str] = None,
        static_dir: Optional[str] = None,
        hot_reload: bool = False,  # whether to use hot reload during debug or not (has no effect in production)
        session_info_extra_content: "Widget" = None,
        session_info_solid: bool = False,
        ready_check_function: Optional[
            Callable
        ] = None,  # function to check if the app is ready for requests (e.g serving app: model is served and ready)
        show_header: bool = True,
    ):
        self._favicon = os.environ.get("icon", "https://cdn.supervisely.com/favicon.ico")
        JinjaWidgets().context["__favicon__"] = self._favicon
        JinjaWidgets().context["__no_html_mode__"] = True
        JinjaWidgets().context["__show_header__"] = show_header

        self._static_dir = static_dir

        self._stop_event = ThreadingEvent()
        # for backward compatibility
        self._graceful_stop_event: Optional[ThreadingEvent] = None
        self.set_ready_check_function(ready_check_function)

        def set_stop_event():
            self._stop_event.set()

        def wait_for_graceful_stop_event():
            if self._graceful_stop_event is None:
                return
            print_info = True
            while not self._graceful_stop_event.is_set():
                if print_info:
                    logger.info("Graceful shutdown. Waiting for `app.stop()` to be called.")
                    print_info = False
                sleep(0.1)

        self._before_shutdown_callbacks = [set_stop_event, wait_for_graceful_stop_event]

        headless = False
        if layout is None and templates_dir is None:
            templates_dir: str = "templates"  # for back compatibility
            headless = True
            logger.info(
                "Both arguments 'layout' and 'templates_dir' are not defined. App is headless (i.e. without UI)",
                extra={"templates_dir": templates_dir},
            )
        if layout is not None and templates_dir is not None:
            raise ValueError(
                "Only one of the arguments has to be defined: 'layout' or 'templates_dir'. 'layout' argument is recommended."
            )
        if layout is not None:
            templates_dir = os.path.join(Path(__file__).parent.absolute(), "templates")
            from supervisely.app.widgets import Identity

            main_layout = Identity(layout, widget_id="__app_main_layout__")
            logger.info(
                "Application is running in no-html mode", extra={"templates_dir": templates_dir}
            )
        else:
            JinjaWidgets().auto_widget_id = False
            JinjaWidgets().context["__no_html_mode__"] = False
        if session_info_extra_content is not None:
            session_info_extra_content_layout = Identity(
                session_info_extra_content, widget_id="__app_session_info_extra_content__"
            )

        if session_info_solid:
            JinjaWidgets().context["__app_session_info_solid__"] = True

        if is_production():
            logger.info("Application is running on Supervisely Platform in production mode")
        else:
            logger.info("Application is running on localhost in development mode")

        self._process_id = os.getpid()
        logger.info(f"Application PID is {self._process_id}")
        self._fastapi: FastAPI = _init(
            app=None,
            templates_dir=templates_dir,
            headless=headless,
            process_id=self._process_id,
            static_dir=static_dir,
            hot_reload=hot_reload,
            before_shutdown_callbacks=self._before_shutdown_callbacks,
        )
        self.test_client = TestClient(self._fastapi)

        if not headless:
            if is_development() and hot_reload:
                templates = Jinja2Templates()
                self.hot_reload = arel.HotReload([])
                self._fastapi.add_websocket_route(
                    "/hot-reload", route=self.hot_reload, name="hot-reload"
                )
                self._fastapi.add_event_handler("startup", self.hot_reload.startup)
                self._fastapi.add_event_handler("shutdown", self.hot_reload.shutdown)

                # Setting HOTRELOAD=1 in template context, otherwise the HTML would not have the hot reload script.
                templates.env.globals["HOTRELOAD"] = "1"
                templates.env.globals["hot_reload"] = self.hot_reload

                logger.debug("Hot reload is enabled, use app.reload_page() to reload page.")

            if is_production():
                # to save offline session
                from supervisely.app.content import ContentOrigin

                ContentOrigin().start()
                Thread(target=run_sync, args=(self.test_client.get("/"),)).start()

        server = self.get_server()

        @server.get("/livez")
        @server.get("/is_alive")
        @server.post("/is_running")
        async def is_running(request: Request):
            is_running = True
            if is_production():
                # @TODO: set task status to running
                return {"running": is_running, "mode": "production"}
            else:
                return {"running": is_running, "mode": "development"}

        @server.get("/readyz")
        @server.get("/is_ready")
        @server.post("/is_ready")
        async def is_ready(response: Response, request: Request):
            is_ready = True
            if self._ready_check_function is not None:
                is_ready = self._ready_check_function()
            if is_ready is False:
                raise HTTPException(status_code=503, detail="Service not ready")
            return {"status": "ready"}

    def get_server(self):
        return self._fastapi

    async def __call__(self, scope, receive, send) -> None:
        await self._fastapi.__call__(scope, receive, send)

    def shutdown(self):
        shutdown(self._process_id, self._before_shutdown_callbacks)

    def stop(self):
        if self._graceful_stop_event is not None:
            self._graceful_stop_event.set()
        if self.is_stopped():
            return
        self.shutdown()

    def is_stopped(self):
        """Indicates whether the application is in the process of being terminated."""
        return self._stop_event.is_set()

    def reload_page(self):
        run_sync(self.hot_reload.notify.notify())

    def get_static_dir(self):
        return self._static_dir

    def call_before_shutdown(self, func: Callable[[], None]):
        self._before_shutdown_callbacks.append(func)

    def handle_stop(self, graceful: bool = True):
        """Contextmanager to suppress StopException and control graceful shutdown.

        :param graceful: Whether to perform a graceful shutdown if a StopException is raised.
        If set to `False` and shutdown request recieved (i.e. `app.is_stopped()` is `True`),
        the application will be terminated immediately, defaults to `True`
        :type graceful: bool
        :return: context manager
        :rtype: _type_
        """
        self._graceful_stop_event = ThreadingEvent()
        if graceful is False:
            self._graceful_stop_event.set()
        return suppress(self.StopException)

    def event(self, event: Event, use_state: bool = False) -> Callable:
        """Decorator to register posts to specific endpoints.
        Supports both async and sync functions.

        :param event: event to register (e.g. `Event.Brush.LeftMouseReleased`)
        :type event: Event
        :param use_state: if set to True, data will be extracted from request.state.state,
            otherwise from request.state.context, defaults to False
        :type use_state: bool, optional
        :return: decorator
        :rtype: Callable

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            app = sly.Application(layout=layout)

            @app.event(sly.Event.Brush.LeftMouseReleased)
            def some_function(api: sly.Api, event: sly.Event.Brush.LeftMouseReleased):
                # do something
                pass
        """

        def inner(func: Callable) -> Callable:
            server = self.get_server()

            if inspect.iscoroutinefunction(func):

                @server.post(event.endpoint)
                async def wrapper(request: Request):
                    data = request.state.state if use_state else request.state.context
                    return await func(request.state.api, event.from_json(data))

            else:

                @server.post(event.endpoint)
                def wrapper(request: Request):
                    data = request.state.state if use_state else request.state.context
                    return func(request.state.api, event.from_json(data))

            return wrapper

        return inner

    def render(self, context: Dict[str, Any]):
        return Jinja2Templates().render("index.html", {**context, "HOTRELOAD": False})

    def set_ready_check_function(self, func: Callable):
        self._ready_check_function = func


def set_autostart_flag_from_state(default: Optional[str] = None):
    """Set `autostart` flag recieved from task state. Env name: `modal.state.autostart`.

    :param default: this value will be set
        if the flag is undefined in state, defaults to None
    :type default: Optional[str], optional
    """
    if sly_env.autostart() is True:
        logger.info("`autostart` flag already defined in env. Skip loading it from state.")
        return

    api = Api()
    task_id = sly_env.task_id(raise_not_found=False)
    if task_id is None:
        logger.warn("`autostart` env can't be setted: TASK_ID variable is not defined.")
        return
    task_meta = api.task.get_info_by_id(task_id).get("meta", None)
    task_params = None
    task_state = None
    auto_start = default
    if task_meta is not None:
        task_params = task_meta.get("params", None)
    if task_params is not None:
        task_state = task_params.get("state", None)
    if task_state is not None:
        auto_start = task_params.get("autostart", default)

    sly_env.set_autostart(auto_start)


def call_on_autostart(
    default_func: Optional[Callable] = None,
    **default_kwargs,
) -> Callable:
    """Decorator to enable autostart.
    This decorator is used to wrap functions that are executed
    and will check if autostart is enabled in environment.

    :param default_func: default function to call if autostart is not enabled, defaults to None
    :type default_func: Optional[Callable], optional
    :return: decorator
    :rtype: Callable
    """
    set_autostart_flag_from_state()

    def inner(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            if sly_env.autostart() is True:
                logger.info("Found `autostart` flag in environment.")
                func(*args, **kwargs)
            else:
                logger.info("Autostart is disabled.")
                if default_func is not None:
                    default_func(**default_kwargs)

        return wrapper

    return inner


def get_name_from_env(default="Supervisely App"):
    return os.environ.get("APP_NAME", default)



from typing import List, Tuple, Union

from supervisely.api.api import Api
from supervisely.api.dataset_api import DatasetInfo
from supervisely.app.widgets import Progress
from supervisely.app.widgets.button.button import Button
from supervisely.app.widgets.checkbox.checkbox import Checkbox
from supervisely.app.widgets.container.container import Container
from supervisely.app.widgets.empty.empty import Empty
from supervisely.app.widgets.flexbox.flexbox import Flexbox
from supervisely.app.widgets.input_number.input_number import InputNumber
from supervisely.app.widgets.notification_box.notification_box import NotificationBox
from supervisely.app.widgets.one_of.one_of import OneOf
from supervisely.app.widgets.project_thumbnail.project_thumbnail import ProjectThumbnail
from supervisely.app.widgets.radio_group.radio_group import RadioGroup
from supervisely.app.widgets.select_dataset.select_dataset import SelectDataset
from supervisely.app.widgets.select_project.select_project import SelectProject
from supervisely.app.widgets.text.text import Text
from supervisely.app.widgets.widget import Widget
from supervisely.project.project import ProjectType
from supervisely.video.sampling import SamplingSettings, sample_video_project


class Sampling(Widget):
    def __init__(
        self,
        project_id: int = None,
        input_selectable: bool = True,
        datasets_ids: List[int] = None,
        output_project_id: int = None,
        output_project_selectable: bool = True,
        widgth: int = 370,
        widget_id: str = None,
        file_path: str = __file__,
    ):
        super().__init__(widget_id, file_path)
        if not input_selectable and project_id is None:
            raise ValueError(
                "Either 'project_id' must be provided or 'project_selectable' must be True."
            )
        if project_id is None and not input_selectable:
            input_selectable = True
        if not output_project_selectable and output_project_id is None:
            raise ValueError(
                "Either 'output_project_id' must be provided or 'output_project_selectable' must be True."
            )
        if output_project_id is None and not output_project_selectable:
            output_project_selectable = True
        self._api = Api()
        self.project_id = project_id
        self.project_selectable = input_selectable
        self.datasets_ids = datasets_ids
        self.output_project_id = output_project_id
        self.output_project_selectable = output_project_selectable
        self.widgth = widgth
        self.project_info = (
            self._api.project.get_info_by_id(self.project_id) if self.project_id else None
        )
        self.all_datasets = (
            self._api.dataset.get_list(self.project_id, recursive=True) if self.project_id else []
        )
        self.items_count = self._count_items(self.all_datasets, datasets_ids, with_children=True)
        self._init_gui()

    def _init_input_gui(self):
        self.input_datasets_select = SelectDataset(
            default_id=self.datasets_ids,
            project_id=self.project_id,
            allowed_project_types=[ProjectType.VIDEOS],
            size="mini",
            multiselect=True,
            select_all_datasets=self.datasets_ids is None,
        )
        self.project_preview = ProjectThumbnail()
        self.project_preview.hide()
        if not self.project_selectable:
            self.input_datasets_select.hide()
            project_info = self._api.project.get_info_by_id(self.project_id)
            self.project_preview.set(project_info)
            self.project_preview.show()
        self.nested_datasets_checkbox = Checkbox(
            "Include nested datasets",
            checked=True,
        )
        self.input_project_container = Container(
            widgets=[
                self.input_datasets_select,
                self.nested_datasets_checkbox,
                self.project_preview,
            ],
            style="padding-left: 21px; padding-top: 10px;",
        )
        self.input_field = Container(
            widgets=[
                Text(
                    '<i class="zmdi zmdi-collection-video" style="padding-right: 10px; color: rgb(0, 154, 255);"></i><b>Input project</b>'
                ),
                self.input_project_container,
            ],
            gap=0,
        )

    def _init_settings_gui(self):
        self.only_annotated_checkbox = Checkbox(Text("Only annotated frames", font_size=13))
        self.only_annotated_row = Flexbox(widgets=[self.only_annotated_checkbox])
        self.step_label = Text("Step:", font_size=13)
        self.step_input = InputNumber(value=1, min=1, step=1, size="mini", width=160)
        self.step_row = Container(
            widgets=[self.step_label, self.step_input],
            direction="horizontal",
            style="width: 202px; align-items: center;",
            widgets_style="flex: 1; display: flex;",
        )
        self.resize_checkbox = Checkbox(Text("Resize frames", font_size=13))
        self.resize_input_h_label = Text("Height:", font_size=13)
        self.resize_input_h = InputNumber(value=224, min=1, step=1, size="mini", controls=False)
        self.resize_input_w_label = Text("Width:", font_size=13)
        self.resize_input_w = InputNumber(value=224, min=1, step=1, size="mini", controls=False)
        self.resize_h_row = Flexbox(widgets=[self.resize_input_h_label, self.resize_input_h])
        self.resize_w_row = Flexbox(
            widgets=[self.resize_input_w_label, self.resize_input_w], gap=15
        )
        self.resize_row = Flexbox(widgets=[self.resize_checkbox])
        self.resize_hw_container = Container(
            widgets=[Empty(), self.resize_h_row, self.resize_w_row]
        )
        self.resize_hw_container.hide()
        self.resize_container = Container(
            widgets=[self.resize_row, self.resize_hw_container], gap=0
        )
        self.copy_annotations_checkbox = Checkbox(
            Text("Copy annotations from source project", font_size=13)
        )
        self.copy_annotations_row = Flexbox(widgets=[self.copy_annotations_checkbox])

        @self.resize_checkbox.value_changed
        def on_resize_checkbox_change(checked: bool):
            if checked:
                self.resize_hw_container.show()
            else:
                self.resize_hw_container.hide()

        self.settings_container = Container(
            widgets=[
                Empty(),
                self.step_row,
                self.only_annotated_row,
                self.resize_container,
                self.copy_annotations_row,
                Empty(),
            ],
            style="padding-left: 21px; padding-top: 10px;",
        )
        self.settings_field = Container(
            widgets=[
                Text(
                    '<i class="zmdi zmdi-settings" style="padding-right: 10px; color: rgb(0, 154, 255);"></i><b>Sampling settings</b>'
                ),
                self.settings_container,
            ],
            gap=0,
        )

    def _datasets_to_process(
        self, all_datasets: List[DatasetInfo], datasets_ids: List[int], with_children: bool
    ) -> List[DatasetInfo]:
        if datasets_ids is None:
            return all_datasets.copy()
        datasets = []
        for ds in [ds for ds in all_datasets if ds.id in datasets_ids]:
            datasets.append(ds)
            if with_children:
                children = [child for child in all_datasets if child.parent_id == ds.id]
                if children:
                    datasets.extend(
                        self._datasets_to_process(
                            all_datasets, [child.id for child in children], True
                        )
                    )
        return datasets

    def _count_items(
        self, all_datasets: List[DatasetInfo], datasets_ids: List[int], with_children: bool
    ) -> int:
        return sum(
            ds.items_count
            for ds in self._datasets_to_process(
                all_datasets, datasets_ids, with_children=with_children
            )
        )

    def _selected_text(self, datasets_ids: List[int] = None, with_children: bool = True) -> str:
        red = "#FF6458;"
        blue = "rgb(0, 154, 255);"
        datasets = self._datasets_to_process(
            self.all_datasets, datasets_ids, with_children=with_children
        )
        color = blue if self.items_count > 0 else red
        ds_num = len(datasets)
        return f'Selected <b style="color: {color};">{ds_num}</b> dataset{"s" if ds_num % 10 != 1 else ""} with <b style="color: {color};">{self.items_count}</b> videos'

    def _update_preview(self):
        with_children = self.nested_datasets_checkbox.is_checked()
        self.selected_items_text.text = self._selected_text(
            self.datasets_ids, with_children=with_children
        )
        if self.items_count > 0:
            self.run_button.enable()
        else:
            self.run_button.disable()

    def _datasets_changed(self, datasets_ids: List[int]):
        self.preview_container.loading = True
        self.datasets_ids = datasets_ids
        project_id = self.input_datasets_select.get_selected_project_id()
        with_children = self.nested_datasets_checkbox.is_checked()
        if self.project_id != project_id:
            if project_id is None:
                self.project_id = None
                self.project_info = None
                self.all_datasets = []
            else:
                self.project_id = project_id
                self._api.project.get_info_by_id(project_id)
                self.all_datasets = self._api.dataset.get_list(project_id, recursive=True)
        self.items_count = self._count_items(
            self.all_datasets, datasets_ids, with_children=with_children
        )
        self._update_preview()
        self.preview_container.loading = False

    def _init_peview_gui(self):
        self.selected_items_text = Text("", font_size=13)
        self.run_button = Button("Run", icon="zmdi zmdi-play", button_size="mini")
        self.run_button.disable()

        self.preview_text = Text(
            '<i class="zmdi zmdi-eye" style="padding-right: 10px; color: rgb(0, 154, 255);"></i><b style="font-size: 14px">Preview</b>',
            font_size=13,
        )
        self.preview_field = Container(
            widgets=[
                self.preview_text,
                Container(widgets=[self.selected_items_text], style="padding-left: 21px;"),
            ],
            style="padding-top: 10px;",
        )
        self.run_button_container = Container(
            widgets=[self.run_button],
            direction="horizontal",
            overflow="wrap",
            style="display: flex; justify-content: flex-end;",
            widgets_style="display: flex; flex: none;",
        )
        self.preview_container = Container(
            widgets=[self.preview_field, self.run_button_container],
        )
        self._update_preview()

        @self.input_datasets_select.value_changed
        def on_input_datasets_select_change(datasets_ids: List[int]):
            self._datasets_changed(datasets_ids)

        @self.nested_datasets_checkbox.value_changed
        def on_nested_datasets_checkbox_change(is_checked: bool):
            self._datasets_changed(self.input_datasets_select.get_selected_ids())

        @self.run_button.click
        def on_run_button_click():
            self.run()

    def _init_output_gui(self):
        self.output_project_select = SelectProject(
            default_id=self.output_project_id, allowed_types=[ProjectType.IMAGES], size="mini"
        )
        self.output_project_preview = ProjectThumbnail()
        self.output_project_preview.hide()
        self.output_mode_radio = RadioGroup(
            items=[
                RadioGroup.Item("create", "Create new project", content=Empty()),
                RadioGroup.Item(
                    "merge",
                    "Merge with existing project",
                    content=self.output_project_select,
                ),
            ]
        )
        self.output_merge_project_oneof = OneOf(self.output_mode_radio)
        self.output_project_container = Container(
            widgets=[self.output_mode_radio, self.output_merge_project_oneof],
        )

        self.output_container = Container(
            widgets=[self.output_project_preview, self.output_project_container],
            gap=0,
            style="padding-left: 21px; padding-top: 10px;",
        )
        if not self.output_project_selectable:
            self.output_project_container.hide()
            project_info = self._api.project.get_info_by_id(self.output_project_id)
            self.output_project_preview.set(project_info)
            self.output_project_preview.show()
        if self.output_project_id is not None:
            self.output_mode_radio.set_value("merge")

        self.output_field = Container(
            widgets=[
                Text(
                    '<i class="zmdi zmdi-collection-folder-image" style="padding-right: 10px; color: rgb(0, 154, 255);"></i><b>Output project</b>'
                ),
                self.output_container,
            ],
            gap=0,
        )

    def _init_progress_gui(self):
        self.items_progress = Progress(hide_on_finish=False)
        self.frames_progress = Progress(hide_on_finish=False)
        self.error_notification = NotificationBox(title="Error", description="", box_type="error")
        self.error_notification.hide()
        self.progress_container = Container(widgets=[self.items_progress, self.frames_progress])
        self.progress_container.hide()
        self.result_project_preview = ProjectThumbnail()
        self.result_project_preview_field = Container(
            widgets=[
                Text(
                    '<i class="zmdi zmdi-check-square" style="padding-right: 10px; color: rgb(0, 154, 255);"></i><b>Result project</b>'
                ),
                Container(
                    widgets=[self.result_project_preview],
                    style="padding-left: 21px;",
                ),
            ]
        )
        self.result_project_preview_field.hide()
        self.result_container = Container(
            widgets=[
                self.progress_container,
                self.error_notification,
                self.result_project_preview_field,
            ],
            gap=0,
        )

    def _init_gui(self):
        self._init_input_gui()
        self._init_settings_gui()
        self._init_output_gui()
        self._init_peview_gui()
        self._init_progress_gui()

        self.content = Container(
            widgets=[
                self.input_field,
                self.settings_field,
                self.output_field,
                self.preview_container,
                self.result_container,
            ],
            style=f"width: {self.widgth}px;",
        )

    def _get_settings(self) -> dict:
        settings = {
            SamplingSettings.ONLY_ANNOTATED: self.only_annotated_checkbox.is_checked(),
            SamplingSettings.STEP: self.step_input.get_value(),
            SamplingSettings.RESIZE: None,
            SamplingSettings.COPY_ANNOTATIONS: self.copy_annotations_checkbox.is_checked(),
        }
        if self.resize_checkbox.is_checked():
            settings[SamplingSettings.RESIZE] = [
                self.resize_input_h.get_value(),
                self.resize_input_w.get_value(),
            ]

        return settings

    def _get_dst_project_id(self) -> int:
        if self.output_mode_radio.get_value() == "create":
            return None
        else:
            return self.output_project_select.get_selected_id()

    def run(self):
        self.progress_container.show()
        self.error_notification.hide()
        self.result_project_preview_field.hide()
        try:
            project_id = self.input_datasets_select.get_selected_project_id()
            datasets_ids = self.input_datasets_select.get_selected_ids()
            with_children = self.nested_datasets_checkbox.is_checked()
            selected_datasets_with_children = self._datasets_to_process(
                all_datasets=self.all_datasets,
                datasets_ids=datasets_ids,
                with_children=with_children,
            )
            total_items = sum(ds.items_count for ds in selected_datasets_with_children)
            datasets_ids = [ds.id for ds in selected_datasets_with_children]
            project_info = self.project_info

            if project_info.type != str(ProjectType.VIDEOS):
                raise ValueError(
                    f"Project with ID {self.input_datasets_select.get_selected_id()} is not a video project."
                )
            frames_pbar = self.frames_progress()
            with self.items_progress(
                message=f"Videos progress...",
                total=total_items,
            ) as pbar:
                self.progress_container.show()
                dst_project_info = sample_video_project(
                    api=self._api,
                    project_id=project_id,
                    settings=self._get_settings(),
                    dst_project_id=self._get_dst_project_id(),
                    datasets_ids=datasets_ids,
                    items_progress_cb=pbar.update,
                    video_progress=frames_pbar,
                )
        except Exception as e:
            self.error_notification.set(title="Error", description=str(e))
            self.error_notification.show()
            raise e
        else:
            if self.output_project_selectable:
                dst_project_info = self._api.project.get_info_by_id(dst_project_info.id)
                self.result_project_preview.set(dst_project_info)
                self.result_project_preview_field.show()
            else:
                dst_project_info = self._api.project.get_info_by_id(self.output_project_id)
                self.output_project_preview.set(dst_project_info)
        finally:
            self.progress_container.hide()

    @property
    def selected_project_id(self) -> int:
        if not self.project_selectable:
            return self.project_id
        return self.input_datasets_select.get_selected_project_id()

    @selected_project_id.setter
    def selected_project_id(self, value: int):
        if not self.project_selectable:
            raise ValueError("Project is not selectable.")
        self.input_datasets_select.set_project_id(value)
        self._datasets_changed(self.input_datasets_select.get_selected_ids())

    @property
    def selected_all_datasets(self) -> bool:
        return self.input_datasets_select._all_datasets_checkbox.is_checked()

    @selected_all_datasets.setter
    def selected_all_datasets(self, value: bool):
        if value:
            self.input_datasets_select._all_datasets_checkbox.check()
        else:
            self.input_datasets_select._all_datasets_checkbox.uncheck()
        self._datasets_changed(self.input_datasets_select.get_selected_ids())

    @property
    def selected_datasets_ids(self) -> List[int]:
        return self.input_datasets_select.get_selected_ids()

    @selected_datasets_ids.setter
    def selected_datasets_ids(self, value: List[int]):
        self.input_datasets_select.set_dataset_ids(value)

    @property
    def include_nested_datasets(self) -> bool:
        return self.nested_datasets_checkbox.is_checked()

    @include_nested_datasets.setter
    def include_nested_datasets(self, value: bool):
        if value:
            self.nested_datasets_checkbox.check()
        else:
            self.nested_datasets_checkbox.uncheck()
        self._datasets_changed(self.input_datasets_select.get_selected_ids())

    @property
    def step(self) -> int:
        return self.step_input.get_value()

    @step.setter
    def step(self, value: int):
        self.step_input.value = value

    @property
    def only_annotated(self) -> bool:
        return self.only_annotated_checkbox.is_checked()

    @only_annotated.setter
    def only_annotated(self, value: bool):
        if value:
            self.only_annotated_checkbox.check()
        else:
            self.only_annotated_checkbox.uncheck()

    @property
    def resize(self) -> Union[Tuple[int, int], None]:
        if self.resize_checkbox.is_checked():
            return (self.resize_input_h.get_value(), self.resize_input_w.get_value())
        return None

    @resize.setter
    def resize(self, value: Union[Tuple[int, int], None]):
        if value is None:
            self.resize_checkbox.uncheck()
        else:
            self.resize_checkbox.check()
            self.resize_input_h.value = value[0]
            self.resize_input_w.value = value[1]

    @property
    def copy_annotations(self) -> bool:
        return self.copy_annotations_checkbox.is_checked()

    @copy_annotations.setter
    def copy_annotations(self, value: bool):
        if value:
            self.copy_annotations_checkbox.check()
        else:
            self.copy_annotations_checkbox.uncheck()

    @property
    def selected_output_project_id(self) -> int:
        if not self.output_project_selectable:
            return self.output_project_id
        if self.output_mode_radio.get_value() == "create":
            return None
        return self.output_project_select.get_selected_id()

    @selected_output_project_id.setter
    def selected_output_project_id(self, value: int):
        if not self.output_project_selectable:
            raise ValueError("Output project is not selectable.")
        self.output_project_select.set_project_id(value)
        if value is not None:
            self.output_mode_radio.set_value("merge")
        else:
            self.output_mode_radio.set_value("create")

    def get_json_data(self) -> dict:
        return {}

    def get_json_state(self) -> dict:
        return {}

    def to_html(self):
        return self.content.to_html()



import os
from collections import defaultdict

import supervisely.convert.image.coco.coco_helper as coco_helper
from supervisely import Annotation, Api, ProjectMeta, batched, is_development, logger
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.coco.coco_converter import COCOConverter
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.io.fs import get_file_ext

COCO_ANN_KEYS = ["images", "annotations"]


class FastCOCOConverter(COCOConverter, ImageConverter):

    def __str__(self) -> str:
        return AvailableImageConverters.FAST_COCO

    def validate_format(self) -> bool:
        from pycocotools.coco import COCO  # pylint: disable=import-error

        detected_ann_cnt = 0
        ann_paths = []
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                if ext == self.ann_ext:
                    ann_paths.append(full_path)
                elif self.is_image(full_path):
                    return False

        if len(ann_paths) == 0:
            return False

        # create Items
        self._items = []
        meta = ProjectMeta()
        warnings = defaultdict(list)
        item_names = set()
        for ann_path in ann_paths:
            try:
                with coco_helper.HiddenCocoPrints():
                    coco = COCO(ann_path)
            except:
                continue
            if not all(key in coco.dataset for key in COCO_ANN_KEYS):
                continue
            coco_anns = coco.imgToAnns
            coco_images = coco.imgs
            if len(coco.cats) > 0:
                coco_categories = coco.loadCats(ids=coco.getCatIds())
            else:
                coco_categories = []
            self._coco_categories.extend(coco_categories)
            coco_items = coco_images.items()
            meta = self.generate_meta_from_annotation(coco, meta)
            # create ann dict
            for image_id, image_info in coco_items:
                image_name = image_info.get("file_name", image_info.get("name"))
                if not isinstance(image_name, str):
                    warnings["file_name field is not a string"].append(image_name)
                    continue
                if "/" in image_name:
                    image_name = os.path.basename(image_name)
                image_url = image_info.get(
                    "coco_url",
                    image_info.get(
                        "flickr_url", image_info.get("url", image_info.get("path", None))
                    ),
                )
                width = image_info.get("width")
                height = image_info.get("height")

                coco_ann = coco_anns[image_id]
                if len(coco_ann) == 0 or coco_ann is None or image_name is None:
                    continue
                if image_name in item_names:
                    # * Block to handle the case when there are mixed annotations: caption and segmentations for the same images
                    item = next(item for item in self._items if item.name == image_name)
                    if item.shape == (height, width):
                        item.ann_data.extend(coco_ann)
                        continue
                item = self.Item(image_name) if image_url is None else self.Item(image_url)
                item.name = image_name
                item.ann_data = coco_ann
                item.set_shape((height, width))
                self._items.append(item)
                item_names.add(image_name)
                detected_ann_cnt += len(coco_ann)

        self._meta = meta
        if len(warnings) > 0:
            for warning, failed_items in warnings.items():
                logger.warn(f"{warning}: {failed_items}")
        return detected_ann_cnt > 0

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: dict = None,
        renamed_tags: dict = None,
    ) -> Annotation:
        """Convert to Supervisely format."""
        try:
            ann = coco_helper.create_supervisely_annotation(
                item,
                meta,
                self._coco_categories,
                renamed_classes,
                renamed_tags,
            )
            return ann
        except Exception as e:
            logger.error(
                f"Error during conversion of annotation for image '{item.name}': {repr(e)}"
            )
            return None

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 50,
        log_progress=True,
    ) -> None:
        """Convert COCO annootations to Supervisely and append to dataset images."""

        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)

        existing_images = {img_info.name: img_info for img_info in api.image.get_list(dataset_id)}
        if len(existing_images) == 0:
            raise RuntimeError(
                "Not found images in the dataset. "
                "Please start the import process from a dataset that contains images, "
                "or upload both images and COCO annotations at once."
            )
        if log_progress:
            progress, progress_cb = self.get_progress(
                self.items_count, "Adding COCO annotations..."
            )
        else:
            progress_cb = None

        for batch in batched(self._items, batch_size=batch_size):
            img_ids = []
            anns = []
            for item in batch:
                existing_image = existing_images.get(item.name)
                if existing_image is None:
                    continue
                if item.shape != (existing_image.height, existing_image.width):
                    logger.warn(
                        f"Image '{item.name}' has different shapes in COCO annotation and Supervisely."
                    )
                    continue

                ann = self.to_supervisely(item, meta, renamed_classes, renamed_tags)
                if ann is not None:
                    img_ids.append(existing_image.id)
                    anns.append(ann)

            if len(anns) == len(img_ids):
                existing_anns = api.annotation.download_json_batch(dataset_id, img_ids)
                existing_anns = [Annotation.from_json(json, meta) for json in existing_anns]
                merged_anns = []
                for existing_ann, new_ann in zip(existing_anns, anns):
                    merged_ann = existing_ann.merge(new_ann)
                    merged_anns.append(merged_ann)

                api.annotation.upload_anns(img_ids, merged_anns)

            if log_progress:
                progress_cb(len(batch))

        if log_progress:
            if is_development():
                progress.close()
        logger.info(f"Dataset ID:'{dataset_id}' has been successfully uploaded.")



from supervisely.nn.benchmark.base_visualizer import BaseVisMetrics
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class RenormalizedErrorOverUnion(BaseVisMetrics):

    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            "renormalized_error_over_union",
            "Renormalized Error over Union",
            text=self.vis_texts.markdown_renormalized_error_ou,
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget("intersection_error_over_union", self.get_figure())

    def get_figure(self):
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()

        labels = ["Boundary EoU", "Extent EoU", "Segment EoU"]

        for idx, eval_result in enumerate(self.eval_results, 1):
            model_name = f"[{idx}] {eval_result.short_name}"

            fig.add_trace(
                go.Bar(
                    x=labels,
                    y=[
                        eval_result.mp.boundary_renormed_eou,
                        eval_result.mp.extent_renormed_eou,
                        eval_result.mp.segment_renormed_eou,
                    ],
                    name=model_name,
                    text=[
                        eval_result.mp.boundary_renormed_eou,
                        eval_result.mp.extent_renormed_eou,
                        eval_result.mp.segment_renormed_eou,
                    ],
                    textposition="outside",
                    marker=dict(color=eval_result.color, line=dict(width=0.7)),
                    width=0.4,
                )
            )

        fig.update_traces(hovertemplate="%{x}: %{y:.2f}<extra></extra>")
        fig.update_layout(
            barmode="group",
            bargap=0.15,
            bargroupgap=0.05,
            width=700 if len(self.eval_results) < 4 else 1000,
        )

        return fig



import time
import uuid
from pathlib import Path
from queue import Queue
from threading import Event, Thread
from typing import Any, BinaryIO, Dict, List, Union

import numpy as np
from pydantic import ValidationError

import supervisely.nn.inference.tracking.functional as F
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.label import Geometry, Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.api.video.video_figure_api import FigureInfo
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.helpers import deserialize_geometry
from supervisely.geometry.point import Point
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.imaging import image as sly_image
from supervisely.nn.inference.tracking.base_tracking import BaseTracking
from supervisely.nn.inference.tracking.tracker_interface import (
    TrackerInterface,
    TrackerInterfaceV2,
)
from supervisely.nn.prediction_dto import Prediction, PredictionPoint
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress


class PointTracking(BaseTracking):
    def _deserialize_geometry(self, data: dict):
        geometry_type_str = data["type"]
        geometry_json = data["data"]
        return deserialize_geometry(geometry_type_str, geometry_json)

    def _inference(self, frames: List[np.ndarray], geometries: List[Geometry], settings: dict):
        updated_settings = {
            **self.custom_inference_settings_dict,
            **settings,
        }
        results = [[] for _ in range(len(frames) - 1)]
        for geometry in geometries:
            if isinstance(geometry, Point):
                predictions = self._predict_point_geometries(geometry, frames, updated_settings)
            elif isinstance(geometry, Polygon):
                if len(geometry.interior) > 0:
                    raise ValueError("Can't track polygons with interior.")
                predictions = self._predict_polygon_geometries(
                    geometry,
                    frames,
                    updated_settings,
                )
            elif isinstance(geometry, GraphNodes):
                predictions = self._predict_graph_geometries(
                    geometry,
                    frames,
                    updated_settings,
                )
            elif isinstance(geometry, Polyline):
                predictions = self._predict_polyline_geometries(
                    geometry,
                    frames,
                    updated_settings,
                )
            else:
                raise TypeError(f"Tracking does not work with {geometry.geometry_name()}.")

            for i, prediction in enumerate(predictions):
                results[i].append({"type": geometry.geometry_name(), "data": prediction.to_json()})

        return results

    def _track(self, api: Api, context: dict):
        if self.custom_inference_settings_dict.get("load_all_frames"):
            load_all_frames = True
        else:
            load_all_frames = False
        video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=load_all_frames,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
        )

        range_of_frames = [
            video_interface.frames_indexes[0],
            video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

        api.logger.info("Start tracking.")

        def _upload_loop(q: Queue, stop_event: Event, video_interface: TrackerInterface):
            try:
                while True:
                    items = []
                    while not q.empty():
                        items.append(q.get_nowait())
                    if len(items) > 0:
                        video_interface.add_object_geometries_on_frames(*list(zip(*items)))
                        continue
                    if stop_event.is_set():
                        video_interface._notify(True, task="stop tracking")
                        return
                    time.sleep(1)
            except Exception as e:
                api.logger.error("Error in upload loop: %s", str(e), exc_info=True)
                video_interface._notify(True, task="stop tracking")
                video_interface.global_stop_indicatior = True
                raise

        upload_queue = Queue()
        stop_upload_event = Event()
        Thread(
            target=_upload_loop,
            args=[upload_queue, stop_upload_event, video_interface],
            daemon=True,
        ).start()
        try:
            for _ in video_interface.frames_loader_generator():
                for (fig_id, geom), obj_id in zip(
                    video_interface.geometries.items(),
                    video_interface.object_ids,
                ):
                    if isinstance(geom, Point):
                        geometries = self._predict_point_geometries(
                            geom,
                            video_interface.frames_with_notification,
                        )
                    elif isinstance(geom, Polygon):
                        if len(geom.interior) > 0:
                            stop_upload_event.set()
                            raise ValueError("Can't track polygons with interior.")
                        geometries = self._predict_polygon_geometries(
                            geom,
                            video_interface.frames_with_notification,
                        )
                    elif isinstance(geom, GraphNodes):
                        geometries = self._predict_graph_geometries(
                            geom,
                            video_interface.frames_with_notification,
                        )
                    elif isinstance(geom, Polyline):
                        geometries = self._predict_polyline_geometries(
                            geom,
                            video_interface.frames_with_notification,
                        )
                    else:
                        raise TypeError(f"Tracking does not work with {geom.geometry_name()}.")

                    for frame_idx, geometry in zip(
                        video_interface._cur_frames_indexes[1:], geometries
                    ):
                        upload_queue.put(
                            (
                                geometry,
                                obj_id,
                                frame_idx,
                            )
                        )
                    api.logger.info(f"Object #{obj_id} tracked.")

                    if video_interface.global_stop_indicatior:
                        stop_upload_event.set()
                        return
        except Exception:
            stop_upload_event.set()
            raise
        stop_upload_event.set()

    def _track_async(self, api: Api, context: dict, inference_request_uuid: str):
        inference_request = self._inference_requests[inference_request_uuid]
        tracker_interface = TrackerInterfaceV2(api, context, self.cache)
        progress: Progress = inference_request["progress"]
        frames_count = tracker_interface.frames_count
        figures = tracker_interface.figures
        progress_total = frames_count * len(figures)
        progress.total = progress_total

        def _upload_f(items: List[FigureInfo]):
            with inference_request["lock"]:
                inference_request["pending_results"].extend(items)

        def _notify_f(items: List[FigureInfo]):
            items_by_object_id: Dict[int, List[FigureInfo]] = {}
            for item in items:
                items_by_object_id.setdefault(item.object_id, []).append(item)

            for object_id, object_items in items_by_object_id.items():
                frame_range = [
                    min(item.frame_index for item in object_items),
                    max(item.frame_index for item in object_items),
                ]
                progress.iters_done_report(len(object_items))
                tracker_interface.notify_progress(progress.current, progress.total, frame_range)

        frame_index = tracker_interface.frame_index
        direction_n = tracker_interface.direction_n
        api.logger.info("Start tracking.")
        try:
            with tracker_interface(_upload_f, _notify_f):
                frames = tracker_interface.load_all_frames()
                frames = [frame.image for frame in frames]
                for figure in figures:
                    figure = api.video.figure._convert_json_info(figure)
                    api.logger.info("geometry:", extra={"figure": figure._asdict()})
                    sly_geometry: Rectangle = deserialize_geometry(
                        figure.geometry_type, figure.geometry
                    )
                    api.logger.info("geometry:", extra={"geometry": type(sly_geometry)})
                    if isinstance(sly_geometry, Point):
                        geometries = self._predict_point_geometries(
                            sly_geometry,
                            frames,
                        )
                    elif isinstance(sly_geometry, Polygon):
                        if len(sly_geometry.interior) > 0:
                            raise ValueError("Can't track polygons with interior.")
                        geometries = self._predict_polygon_geometries(
                            sly_geometry,
                            frames,
                        )
                    elif isinstance(sly_geometry, GraphNodes):
                        geometries = self._predict_graph_geometries(
                            sly_geometry,
                            frames,
                        )
                    elif isinstance(sly_geometry, Polyline):
                        geometries = self._predict_polyline_geometries(
                            sly_geometry,
                            frames,
                        )
                    else:
                        raise TypeError(
                            f"Tracking does not work with {sly_geometry.geometry_name()}."
                        )

                    for i, geometry in enumerate(geometries, 1):
                        figure_id = uuid.uuid5(
                            namespace=uuid.NAMESPACE_URL, name=f"{time.time()}"
                        ).hex
                        result_figure = api.video.figure._convert_json_info(
                            {
                                ApiField.ID: figure_id,
                                ApiField.OBJECT_ID: figure.object_id,
                                "meta": {"frame": frame_index + i * direction_n},
                                ApiField.GEOMETRY_TYPE: geometry.geometry_name(),
                                ApiField.GEOMETRY: geometry.to_json(),
                                ApiField.TRACK_ID: tracker_interface.track_id,
                            }
                        )
                        tracker_interface.add_prediction(result_figure)
                    api.logger.info(f"Figure #{figure.id} tracked.")

                    if inference_request["cancel_inference"]:
                        return
                    if tracker_interface.is_stopped():
                        reason = tracker_interface.stop_reason()
                        if isinstance(reason, Exception):
                            raise reason
                        return
        except Exception as e:
            progress.message = "Error occured during tracking"
            raise
        else:
            progress.message = "Ready"
        finally:
            progress.set(current=0, total=1, report=True)

    def track(self, api: Api, state: Dict, context: Dict):
        fn = self.send_error_data(api, context)(self._track)
        self.schedule_task(fn, api, context)
        return {"message": "Track task started."}

    def track_api(self, api: Api, state: Dict, context: Dict):
        # unused fields:
        context["trackId"] = "auto"
        context["objectIds"] = []
        context["figureIds"] = []
        if "direction" not in context:
            context["direction"] = "forward"

        input_geometries: list = context["input_geometries"]

        if self.custom_inference_settings_dict.get("load_all_frames"):
            load_all_frames = True
        else:
            load_all_frames = False
        video_interface = TrackerInterface(
            context=context,
            api=api,
            load_all_frames=load_all_frames,
            frame_loader=self.cache.download_frame,
            frames_loader=self.cache.download_frames,
            should_notify=False,
        )

        range_of_frames = [
            video_interface.frames_indexes[0],
            video_interface.frames_indexes[-1],
        ]

        if self.cache.is_persistent:
            # if cache is persistent, run cache task for whole video
            self.cache.run_cache_task_manually(
                api,
                None,
                video_id=video_interface.video_id,
            )
        else:
            # if cache is not persistent, run cache task for range of frames
            self.cache.run_cache_task_manually(
                api,
                [range_of_frames],
                video_id=video_interface.video_id,
            )

        api.logger.info("Start tracking.")

        predictions = []
        for _ in video_interface.frames_loader_generator():
            for input_geom in input_geometries:
                geom = self._deserialize_geometry(input_geom)
                if isinstance(geom, Point):
                    geometries = self._predict_point_geometries(
                        geom,
                        video_interface.frames,
                    )
                elif isinstance(geom, Polygon):
                    if len(geom.interior) > 0:
                        raise ValueError("Can't track polygons with interior.")
                    geometries = self._predict_polygon_geometries(
                        geom,
                        video_interface.frames,
                    )
                elif isinstance(geom, GraphNodes):
                    geometries = self._predict_graph_geometries(
                        geom,
                        video_interface.frames,
                    )
                elif isinstance(geom, Polyline):
                    geometries = self._predict_polyline_geometries(
                        geom,
                        video_interface.frames,
                    )
                else:
                    raise TypeError(f"Tracking does not work with {geom.geometry_name()}.")

                if video_interface.global_stop_indicatior:
                    return

                geometries = [video_interface._crop_geometry(g) for g in geometries]
                geometries = [g for g in geometries if g is not None]
                geometries = [{"type": g.geometry_name(), "data": g.to_json()} for g in geometries]
                predictions.append(geometries)

        # predictions must be NxK figures: N=number of frames, K=number of objects
        predictions = list(map(list, zip(*predictions)))
        return predictions

    def track_api_files(
        self,
        files: List[BinaryIO],
        settings: Dict,
    ):
        logger.info(f"Start tracking with settings:", extra={"settings": settings})
        frame_indexes = list(
            range(settings["frame_index"], settings["frame_index"] + settings["frames"] + 1)
        )
        geometries = map(self._deserialize_geometry, settings["input_geometries"])
        frames = []
        for file, frame_idx in zip(files, frame_indexes):
            img_bytes = file.read()
            frame = sly_image.read_bytes(img_bytes)
            frames.append(frame)
        logger.info("Start tracking.")
        return self._inference(frames, geometries, settings)

    def track_async(self, api: Api, state: Dict, context: Dict):
        batch_size = context.get("batch_size", self.get_batch_size())
        if self.max_batch_size is not None and batch_size > self.max_batch_size:
            raise ValidationError(
                f"Batch size should be less than or equal to {self.max_batch_size} for this model."
            )

        inference_request_uuid = uuid.uuid5(namespace=uuid.NAMESPACE_URL, name=f"{time.time()}").hex
        fn = self.send_error_data(api, context)(self._track_async)
        self.schedule_task(fn, api, context, inference_request_uuid=inference_request_uuid)

        logger.debug(
            "Inference has scheduled from 'track_async' endpoint",
            extra={"inference_request_uuid": inference_request_uuid},
        )
        return {
            "message": "Inference has started.",
            "inference_request_uuid": inference_request_uuid,
        }

    def predict(
        self,
        rgb_images: List[np.ndarray],
        settings: Dict[str, Any],
        start_object: Union[PredictionPoint, List[PredictionPoint]],
    ) -> List[PredictionPoint]:
        """
        Track point on given frames.

        :param rgb_images: RGB frames, `m` frames
        :type rgb_images: List[np.array]
        :param settings: model parameters
        :type settings: Dict[str, Any]
        :param start_object: point to track on the initial frame
        :type start_object: PredictionPoint
        :return: predicted points for frame range (0, m]; `m-1` prediction in total
        :rtype: List[PredictionPoint]
        """
        raise NotImplementedError

    def predict_batch(
        self,
        rgb_images: List[np.ndarray],
        settings: Dict[str, Any],
        start_objects: List[PredictionPoint],
    ) -> List[List[PredictionPoint]]:
        """
        Track points on given frames.

        :param rgb_images: RGB frames, `m` frames
        :type rgb_images: List[np.array]
        :param settings: model parameters
        :type settings: Dict[str, Any]
        :param start_objects: points to track on the initial frame
        :type start_objects: List[PredictionPoint]
        :return: predicted points for frame range (0, m]; `m-1` prediction in total
        :rtype: List[List[PredictionPoint]]
        """
        raise NotImplementedError

    def visualize(
        self,
        predictions: List[PredictionPoint],
        images: List[np.ndarray],
        vis_path: str,
        thickness: int = 2,
        classes_whitelist: List[str] = None,
    ):
        vis_path = Path(vis_path)

        for i, (pred, image) in enumerate(zip(predictions, images)):
            out_path = vis_path / f"img_{i}.jpg"
            ann = self._predictions_to_annotation(image, [pred], classes_whitelist)
            ann.draw_pretty(
                bitmap=image,
                color=(255, 0, 0),
                thickness=thickness,
                output_path=str(out_path),
                fill_rectangles=False,
            )

    def _create_label(self, dto: PredictionPoint) -> Point:
        geometry = Point(row=dto.row, col=dto.col)
        return Label(geometry, ObjClass("", Point))

    def _get_obj_class_shape(self):
        return Point

    def _predict_point_geometries(
        self,
        geom: Point,
        frames: List[np.ndarray],
        settings: Dict[str, Any] = None,
    ) -> List[Point]:
        if settings is None:
            settings = self.custom_inference_settings_dict
        pp_geom = PredictionPoint("point", col=geom.col, row=geom.row)
        if type(self).predict_batch == PointTracking.predict_batch:
            # if predict_batch is not implemented, we can't use it
            predicted = self.predict(
                frames,
                settings,
                pp_geom,
            )
        else:
            predicted = self.predict_batch(
                frames,
                settings,
                pp_geom,
            )
            predicted = [pred[0] for pred in predicted]
        return F.dto_points_to_sly_points(predicted)

    def _predict_polygon_geometries(
        self,
        geom: Polygon,
        frames: List[np.ndarray],
        settings: Dict[str, Any] = None,
    ) -> List[Polygon]:
        if settings is None:
            settings = self.custom_inference_settings_dict
        polygon_points = F.numpy_to_dto_point(geom.exterior_np, "polygon")

        if type(self).predict_batch == PointTracking.predict_batch:
            # if predict_batch is not implemented, we can't use it
            points = list(
                zip(
                    *[
                        self.predict(frames, settings, polygon_point)
                        for polygon_point in polygon_points
                    ]
                )
            )
        else:
            points: List[List[Prediction]] = self.predict_batch(
                frames,
                settings,
                polygon_points,
            )
        points_loc = [F.dto_points_to_point_location(frame_points) for frame_points in points]
        return F.exteriors_to_sly_polygons(points_loc)

    def _predict_graph_geometries(
        self,
        geom: GraphNodes,
        frames: List[np.ndarray],
        settings: Dict[str, Any] = None,
    ) -> List[GraphNodes]:
        if settings is None:
            settings = self.custom_inference_settings_dict
        points, pids = F.graph_to_dto_points(geom)

        if type(self).predict_batch == PointTracking.predict_batch:
            # if predict_batch is not implemented, we can't use it
            preds = list(zip(*[self.predict(frames, settings, point) for point in points]))
        else:
            preds: List[List[PredictionPoint]] = self.predict_batch(
                frames,
                settings,
                points,
            )

        nodes = []
        for frame_preds in preds:
            frame_nodes = []
            for pred, pid in zip(frame_preds, pids):
                frame_nodes.extend(F.dto_points_to_sly_nodes([pred], pid))
            nodes.append(frame_nodes)

        return F.nodes_to_sly_graph(nodes)

    def _predict_polyline_geometries(
        self,
        geom: Polyline,
        frames: List[np.ndarray],
        settings: Dict[str, Any] = None,
    ) -> List[Polyline]:
        if settings is None:
            settings = self.custom_inference_settings_dict
        polyline_points = F.numpy_to_dto_point(geom.exterior_np, "polyline")
        if type(self).predict_batch == PointTracking.predict_batch:
            # if predict_batch is not implemented, we can't use it
            preds = list(
                zip(
                    *[
                        self.predict(frames, settings, polyline_point)
                        for polyline_point in polyline_points
                    ]
                )
            )
        else:
            preds = self.predict_batch(
                frames,
                settings,
                polyline_points,
            )
        points_loc = [F.dto_points_to_point_location(frame_points) for frame_points in preds]
        return F.exterior_to_sly_polyline(points_loc)

    def _predictions_to_annotation(
        self, image: np.ndarray, predictions: List[Prediction], classes_whitelist: List[str] = None
    ) -> Annotation:
        labels = []
        for prediction in predictions:
            if (
                not classes_whitelist in (None, "all")
                and prediction.class_name not in classes_whitelist
            ):
                continue
            label = self._create_label(prediction)
            if label is None:
                # for example empty mask
                continue
            if isinstance(label, list):
                labels.extend(label)
                continue
            labels.append(label)

        # create annotation with correct image resolution
        ann = Annotation(img_size=image.shape[:2])
        ann = ann.add_labels(labels)
        return ann



# coding: utf-8

# docs
from supervisely.pointcloud_annotation.pointcloud_episode_tag_collection import (
    PointcloudEpisodeTagCollection,
)
from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject


class PointcloudEpisodeObject(PointcloudObject):
    """
    PointcloudEpisodeObject object for :class:`PointcloudEpisodeAnnotation<supervisely.pointcloud_annotation.pointcloud_episode_annotation.PointcloudEpisodeAnnotation>`. :class:`PointcloudEpisodeObject<PointcloudEpisodeObject>` object is immutable.

    :param obj_class: :class:`class<supervisely.annotation.obj_class.ObjClass>` object.
    :type obj_class: ObjClass
    :param tags: :class:`PointcloudEpisodeTagCollection<supervisely.pointcloud_annotation.pointcloud_episode_tag_collection.PointcloudEpisodeTagCollection>` object.
    :type tags: PointcloudEpisodeTagCollection, optional
    :param key: KeyIdMap object.
    :type key: KeyIdMap, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which PointcloudEpisodeObject belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created PointcloudEpisodeObject.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when PointcloudEpisodeObject was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when PointcloudEpisodeObject was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :Usage example:

     .. code-block:: python

        import supervisely as sly
        from supervisely.geometry.cuboid_3d import Cuboid3d

        obj_class_car = sly.ObjClass('car', Cuboid3d)
        pointcloud_episode_obj_car = sly.PointcloudEpisodeObject(obj_class_car)
        pointcloud_episode_obj_car_json = pointcloud_episode_obj_car.to_json()
        print(pointcloud_episode_obj_car_json)
        # Output: {
        #     "key": "6b819f1840f84d669b32cdec225385f0",
        #     "classTitle": "car",
        #     "tags": []
        # }
    """

    tag_collection_type = PointcloudEpisodeTagCollection



import os
from typing import Dict, Optional

import supervisely.convert.image.sly.sly_image_helper as sly_image_helper
from supervisely import (
    Annotation,
    Dataset,
    Label,
    OpenMode,
    Project,
    ProjectMeta,
    Rectangle,
    logger,
)
from supervisely._utils import generate_free_name, is_development
from supervisely.api.api import Api
from supervisely.convert.base_converter import AvailableImageConverters
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.convert.image.image_helper import validate_image_bounds
from supervisely.io.fs import dirs_filter, file_exists, get_file_ext, get_file_name
from supervisely.io.json import load_json_file
from supervisely.project.project import find_project_dirs
from supervisely.project.project import upload_project as upload_project_fs
from supervisely.project.project_settings import LabelingInterface

DATASET_ITEMS = "items"
NESTED_DATASETS = "datasets"


class SLYImageConverter(ImageConverter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._project_structure = None
        self._supports_links = True
        self._blob_project = False

    def __str__(self):
        return AvailableImageConverters.SLY

    @property
    def blob_project(self) -> bool:
        return self._blob_project

    @blob_project.setter
    def blob_project(self, value: bool):
        self._blob_project = value

    @property
    def ann_ext(self) -> str:
        return ".json"

    @property
    def key_file_ext(self) -> str:
        return ".json"

    def validate_labeling_interface(self) -> bool:
        return self._labeling_interface in [
            LabelingInterface.DEFAULT,
            LabelingInterface.IMAGE_MATTING,
            LabelingInterface.FISHEYE,
            LabelingInterface.MULTIVIEW,
        ]

    def generate_meta_from_annotation(self, ann_path: str, meta: ProjectMeta) -> ProjectMeta:
        ann_json = load_json_file(ann_path)
        meta = sly_image_helper.get_meta_from_annotation(ann_json, meta)
        return meta

    def validate_ann_file(self, ann_path: str, meta: ProjectMeta) -> bool:
        try:
            ann_json = load_json_file(ann_path)
            if "annotation" in ann_json:
                ann_json = ann_json["annotation"]
            ann = Annotation.from_json(ann_json, meta)
            return True
        except:
            return False

    def validate_key_file(self, key_file_path: str) -> bool:
        try:
            self._meta = ProjectMeta.from_json(load_json_file(key_file_path))
            return True
        except Exception:
            return False

    def validate_format(self) -> bool:
        if self.upload_as_links and self._supports_links:
            self._download_remote_ann_files()
        if self.read_sly_project(self._input_data):
            return True

        if self.read_sly_dataset(self._input_data):
            return True

        detected_ann_cnt = 0
        images_list, ann_dict, img_meta_dict = [], {}, {}
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                dir_name = os.path.basename(root)
                if file == "meta.json":
                    is_valid = self.validate_key_file(full_path)
                    if is_valid:
                        continue

                ext = get_file_ext(full_path)
                if ext in self.ann_ext:
                    if dir_name == "meta":
                        img_meta_dict[file] = full_path
                    else:
                        ann_dict[file] = full_path
                elif self.is_image(full_path):
                    images_list.append(full_path)

        if self._meta is not None:
            meta = self._meta
        else:
            meta = ProjectMeta()

        # create Items
        self._items = []
        for image_path in images_list:
            item = self.Item(image_path)
            json_name, json_name_noext = f"{item.name}.json", f"{get_file_name(item.name)}.json"
            ann_path = ann_dict.get(json_name, ann_dict.get(json_name_noext))
            if ann_path:
                if self._meta is None:
                    meta = self.generate_meta_from_annotation(ann_path, meta)
                is_valid = self.validate_ann_file(ann_path, meta)
                if is_valid:
                    item.ann_data = ann_path
            
            meta_path = img_meta_dict.get(json_name, img_meta_dict.get(json_name_noext))
            if meta_path:
                item.set_meta_data(meta_path)
            if item.ann_data is not None or item.meta is not None:
                detected_ann_cnt += 1
            self._items.append(item)
        self._meta = meta
        return detected_ann_cnt > 0

    def to_supervisely(
        self,
        item: ImageConverter.Item,
        meta: ProjectMeta = None,
        renamed_classes: Optional[Dict[str, str]] = None,
        renamed_tags: Optional[Dict[str, str]] = None,
    ) -> Annotation:
        """Convert to Supervisely format."""
        if meta is None:
            meta = self._meta

        if item.ann_data is None:
            if self._upload_as_links:
                item.set_shape([None, None])
            return item.create_empty_annotation()

        try:
            ann_json = load_json_file(item.ann_data)
            if "annotation" in ann_json:
                ann_json = ann_json["annotation"]
            if renamed_classes or renamed_tags:
                ann_json = sly_image_helper.rename_in_json(ann_json, renamed_classes, renamed_tags)
            img_size = list(ann_json["size"].values())
            labels = validate_image_bounds(
                [Label.from_json(obj, meta) for obj in ann_json["objects"]],
                Rectangle.from_size(img_size),
            )
            return Annotation.from_json(ann_json, meta).clone(labels=labels)
        except Exception as e:
            logger.warning(f"Failed to convert annotation: {repr(e)}")
            return item.create_empty_annotation()

    def read_sly_project(self, input_data: str) -> bool:
        try:
            self._items = []
            project = {}
            ds_cnt = 0
            self._meta = None
            logger.debug("Trying to find Supervisely project format in the input data")
            project_dirs = [d for d in find_project_dirs(input_data)]
            if len(project_dirs) > 1:
                logger.info("Found multiple possible Supervisely projects in the input data")
            elif len(project_dirs) == 1:
                logger.info("Possible Supervisely project found in the input data")
            meta = None
            for project_dir in project_dirs:
                project_fs = Project(project_dir, mode=OpenMode.READ)
                if len(project_fs.blob_files) > 0:
                    self.blob_project = True
                    logger.info("Found blob files in the project, skipping")
                    continue

                if meta is None:
                    meta = project_fs.meta
                else:
                    meta = meta.merge(project_fs.meta)
                for dataset in project_fs.datasets:
                    ds_items = []
                    for name in dataset.get_items_names():
                        img_path, ann_path = dataset.get_item_paths(name)
                        meta_path = dataset.get_item_meta_path(name)
                        item = self.Item(img_path)
                        if file_exists(ann_path):
                            if self.validate_ann_file(ann_path, meta):
                                item.ann_data = ann_path
                        if file_exists(meta_path):
                            item.set_meta_data(meta_path)
                        ds_items.append(item)
                    if len(ds_items) > 0:
                        parts = dataset.name.split("/")
                        curr_ds = project.setdefault(
                            parts[0], {DATASET_ITEMS: [], NESTED_DATASETS: {}}
                        )
                        for part in parts[1:]:
                            curr_ds = curr_ds[NESTED_DATASETS].setdefault(
                                part, {DATASET_ITEMS: [], NESTED_DATASETS: {}}
                            )
                        curr_ds[DATASET_ITEMS].extend(ds_items)
                        ds_cnt += 1
                        self._items.extend(ds_items)
            if self.items_count > 0:
                self._meta = meta
                if ds_cnt > 1:  # multiple datasets
                    self._project_structure = project
                return True
            elif self.blob_project:
                return True
            else:
                return False
        except Exception as e:
            logger.debug(f"Not a Supervisely project: {repr(e)}")
            return False

    def read_sly_dataset(self, input_data: str) -> bool:
        try:
            self._items = []
            project = {}
            ds_cnt = 0
            self._meta = None
            logger.debug("Trying to read Supervisely datasets")

            def _check_function(path):
                try:
                    dataset_ds = Dataset(path, OpenMode.READ)
                    return len(dataset_ds.get_items_names()) > 0
                except:
                    return False

            meta = ProjectMeta()
            dataset_dirs = [d for d in dirs_filter(input_data, _check_function)]
            for dataset_dir in dataset_dirs:
                dataset_fs = Dataset(dataset_dir, OpenMode.READ)
                ds_items = []
                for name in dataset_fs.get_items_names():
                    img_path, ann_path = dataset_fs.get_item_paths(name)
                    meta_path = dataset_fs.get_item_meta_path(name)
                    item = self.Item(img_path)
                    if file_exists(ann_path):
                        meta = self.generate_meta_from_annotation(ann_path, meta)
                        if self.validate_ann_file(ann_path, meta):
                            item.ann_data = ann_path
                    if file_exists(meta_path):
                        item.set_meta_data(meta_path)
                    ds_items.append(item)
                if len(ds_items) > 0:
                    parts = dataset_fs.name.split("/")
                    curr_ds = project.setdefault(parts[0], {DATASET_ITEMS: [], NESTED_DATASETS: {}})
                    for part in parts[1:]:
                        curr_ds = curr_ds[NESTED_DATASETS].setdefault(
                            part, {DATASET_ITEMS: [], NESTED_DATASETS: {}}
                        )
                    curr_ds[DATASET_ITEMS].extend(ds_items)
                    ds_cnt += 1
                    self._items.extend(ds_items)

            if self.items_count > 0:
                self._meta = meta
                if ds_cnt > 1:  # multiple datasets
                    self._project_structure = project
                return True
            else:
                return False
        except Exception as e:
            logger.debug(f"Failed to read Supervisely datasets: {repr(e)}")
            return False

    def upload_dataset(
        self, api: Api, dataset_id: int, batch_size: int = 50, log_progress=True
    ) -> None:

        if self._project_structure:
            self.upload_project(api, dataset_id, batch_size, log_progress)
        elif self.blob_project:
            dataset_info = api.dataset.get_info_by_id(dataset_id, raise_error=True)
            project_dirs = [d for d in find_project_dirs(self._input_data)]
            if len(project_dirs) == 0:
                raise RuntimeError(
                    "Failed to find Supervisely project with blobs in the input data"
                )
            project_dir = project_dirs[0]
            if len(project_dirs) > 1:
                logger.info(
                    "Found multiple possible Supervisely projects with blobs in the input data. "
                    f"Only the first one will be uploaded: {project_dir}"
                )
            upload_project_fs(
                dir=project_dir,
                api=api,
                workspace_id=dataset_info.workspace_id,
                log_progress=log_progress,
                project_id=dataset_info.project_id,
            )
        else:
            super().upload_dataset(api, dataset_id, batch_size, log_progress)

    def upload_project(
        self, api: Api, dataset_id: int, batch_size: int = 50, log_progress=True
    ) -> None:
        dataset_info = api.dataset.get_info_by_id(dataset_id, raise_error=True)
        project_id = dataset_info.project_id
        existing_datasets = api.dataset.get_list(project_id, recursive=True)
        existing_datasets = {ds.name for ds in existing_datasets}

        if log_progress:
            progress, progress_cb = self.get_progress(self.items_count, "Uploading project")
        else:
            progress, progress_cb = None, None

        logger.info("Uploading project structure")

        def _upload_project(
            project_structure: Dict,
            project_id: int,
            dataset_id: int,
            parent_id: Optional[int] = None,
            first_dataset=False,
        ):
            for ds_name, value in project_structure.items():
                ds_name = generate_free_name(existing_datasets, ds_name, extend_used_names=True)
                if first_dataset:
                    first_dataset = False
                    api.dataset.update(dataset_id, ds_name)  # rename first dataset
                else:
                    dataset_id = api.dataset.create(project_id, ds_name, parent_id=parent_id).id

                items = value.get(DATASET_ITEMS, [])
                nested_datasets = value.get(NESTED_DATASETS, {})
                logger.info(
                    f"Dataset: {ds_name}, items: {len(items)}, nested datasets: {len(nested_datasets)}"
                )
                if items:
                    super(SLYImageConverter, self).upload_dataset(
                        api, dataset_id, batch_size, entities=items, progress_cb=progress_cb
                    )

                if nested_datasets:
                    _upload_project(nested_datasets, project_id, dataset_id, dataset_id)

        _upload_project(self._project_structure, project_id, dataset_id, first_dataset=True)

        if is_development() and progress is not None:
            progress.close()
